import hashlib
from typing import Iterable, TYPE_CHECKING

from ravendb.primitives import constants
from ravendb.tools.utils import Utils

if TYPE_CHECKING:
    from ravendb.documents.queries.index_query import Parameters


class HashCalculator:
    def __init__(self):
        self.__buffer = []

    @property
    def hash(self) -> str:
        return self.flush_md5()

    @staticmethod
    def __convert_to_hashable(list_):
        for item in list_:
            if isinstance(item, list):
                HashCalculator.__convert_to_hashable(item)
        return tuple(list_)

    def write(self, obj: object):
        if obj is None:
            self.__buffer.append("null-object")
        elif isinstance(obj, (bool, int, float, str, bytes, bytearray)):
            self.__buffer.append(str(obj))
        elif isinstance(obj, list):
            self.__buffer.append(str(self.__convert_to_hashable(obj)))
        elif isinstance(obj, dict):
            self.__buffer.append(str(self.__convert_to_hashable(obj.items())))
        elif "__str__" in obj.__class__.__dict__:
            self.__buffer.append(str(obj))
        else:
            self.__buffer.append(str(Utils.dictionarize(obj)))

    def write_parameters(self, qp: "Parameters") -> None:
        if qp is None:
            self.write("null-params")
        else:
            self.write(len(qp))
            for key, value in qp.items():
                self.__write_parameter_value(value)

    def __write_parameter_value(self, value: object) -> None:
        self.write(value)

    def flush_md5(self) -> str:
        return self.calculate_hash_from_str_collection(self.__buffer)

    @staticmethod
    def calculate_hash_from_str_collection(unique_ids: Iterable[str]) -> str:
        return hashlib.md5(str(hash(tuple(unique_ids))).encode("utf-8")).hexdigest()


class QueryFieldUtil:
    @staticmethod
    def __should_escape(name: str, is_path: bool) -> bool:
        escape = False
        inside_escaped = False
        first = True

        for c in name:
            if c == "'" or c == '"':
                inside_escaped = not inside_escaped
                continue

            if first:
                if not c.isalpha() and c != "_" and c != "@" and not inside_escaped:
                    escape = True
                    break
                first = False
            else:
                if (
                    (not c.isalpha() and not c.isdigit())
                    and c != "_"
                    and c != "@"
                    and c != "["
                    and c != "]"
                    and c != "("
                    and c != ")"
                    and c != "."
                    and not inside_escaped
                ):
                    escape = True
                    break

                if is_path and c == "." and not inside_escaped:
                    escape = True
                    break
        escape |= inside_escaped
        return escape

    @staticmethod
    def escape_if_necessary(name: str, is_path: bool = False) -> str:
        if (
            not name
            or name == constants.Documents.Indexing.Fields.DOCUMENT_ID_FIELD_NAME
            or name == constants.Documents.Indexing.Fields.REDUCE_KEY_HASH_FIELD_NAME
            or name == constants.Documents.Indexing.Fields.REDUCE_KEY_KEY_VALUE_FIELD_NAME
            or name == constants.Documents.Indexing.Fields.VALUE_FIELD_NAME
            or name == constants.Documents.Indexing.Fields.SPATIAL_SHAPE_FIELD_NAME
        ):
            return name

        if not QueryFieldUtil.__should_escape(name, is_path):
            return name

        sb = [name]
        need_end_quote = False
        last_term_start = 0
        offset = 0
        for i in range(len(sb)):
            if i + offset == len(sb):
                break
            c = sb[i + offset]
            if i + offset == 0 and not c.isalpha() and c != "_" and c != "@":
                sb.insert(last_term_start, "'")
                need_end_quote = True
                continue

            if is_path and c == ".":
                if need_end_quote:
                    need_end_quote = False
                    sb.insert(i + offset, "'")
                    offset += 1

                last_term_start = i + offset + 1
                continue

            if not c.isalnum() and c not in ["_", "-", "@", ".", "[", "]"] and not need_end_quote:
                sb.insert(last_term_start, "'")
                need_end_quote = True
                continue

        if need_end_quote:
            sb.append("'")

        return "".join(sb)



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Union, Set, Dict, Callable, Optional, List, TYPE_CHECKING

from ravendb.documents.queries.facets.misc import FacetOptions, FacetAggregation
from ravendb.documents.session.tokens.query_tokens.facets import FacetToken

if TYPE_CHECKING:
    from ravendb.documents.queries.facets.builders import RangeBuilder


class FacetAggregationField:
    def __init__(self, name: str = None, display_name: str = None):
        self.name = name
        self.display_name = display_name

    def __eq__(self, o):
        if self == o:
            return True

        if o is None or self.__class__ != type(o):
            return False

        return self.name == o.name and self.display_name == o.display_name

    def __hash__(self):
        return hash((self.name, self.display_name))

    def to_json(self) -> Dict:
        return {"Name": self.name, "DisplayName": self.display_name}

    @classmethod
    def from_json(cls, json_dict: Dict) -> FacetAggregationField:
        return cls(json_dict["Name"], json_dict["DisplayName"])


class FacetBase(ABC):
    def __init__(self):
        self.display_field_name: Union[None, str] = None
        self.options: Union[None, FacetOptions] = None
        self.aggregations: Dict[FacetAggregation, Set[FacetAggregationField]] = {}

    def to_facet_token(self, add_query_parameter: Callable[[object], str]) -> FacetToken:
        pass

    def to_json(self) -> Dict:
        json_dict = {
            "DisplayFieldName": self.display_field_name,
            "Aggregations": {
                aggregation.to_json(): {aggregation_field.to_json() for aggregation_field in aggregation_fields_set}
                for aggregation, aggregation_fields_set in self.aggregations.items()
            },
        }
        if self.options:
            json_dict["Options"] = self.options.to_json()
        return json_dict

    @classmethod
    def from_json(cls, json_dict: Dict) -> FacetBase:
        facet_base = cls()
        facet_base.display_field_name = json_dict["DisplayFieldName"]
        facet_base.options = FacetOptions.from_json(json_dict["Options"])
        facet_base.aggregations = {
            FacetAggregation.from_json(aggregation): {
                FacetAggregationField.from_json(field) for field in aggregation_fields
            }
            for aggregation, aggregation_fields in json_dict["Aggregations"]
        }
        return facet_base


class GenericRangeFacet(FacetBase):
    def __init__(self, parent: Optional[FacetBase] = None):
        super().__init__()
        self.ranges: List[RangeBuilder] = []
        self.__parent = parent

    @staticmethod
    def parse(range_builder: "RangeBuilder", add_query_parameter: Callable[[object], str]) -> str:
        return range_builder.get_string_representation(add_query_parameter)

    def to_facet_token(self, add_query_parameter: Callable[[object], str]) -> FacetToken:
        if self.__parent is not None:
            return self.__parent.to_facet_token(add_query_parameter)

        return FacetToken.create_from_generic_range_facet(self, add_query_parameter)


class RangeFacet(FacetBase):
    def __init__(self, parent: Optional[FacetBase] = None):
        super().__init__()
        self.ranges: List[str] = []
        self.__parent = parent

    @classmethod
    def from_json(cls, json_dict: Dict) -> RangeFacet:
        range_facet = cls(FacetBase.from_json(json_dict))
        range_facet.aggregations = range_facet.__parent.aggregations
        range_facet.options = range_facet.__parent.options
        range_facet.display_field_name = range_facet.__parent.display_field_name
        range_facet.ranges = json_dict["Ranges"]
        return range_facet

    def to_json(self) -> Dict:
        super_json = super().to_json()
        super_json.update({"Ranges": self.ranges})
        return super_json

    def to_facet_token(self, add_query_parameter: Callable[[object], str]) -> FacetToken:
        if self.__parent is not None:
            return self.__parent.to_facet_token(add_query_parameter)

        return FacetToken.create_from_range_facet(self, add_query_parameter)


class Facet(FacetBase):
    def __init__(self, field_name: str = None):
        super().__init__()
        self.field_name = field_name

    def to_facet_token(self, add_query_parameter: Callable[[object], str]) -> FacetToken:
        return FacetToken.create_from_facet(self, add_query_parameter)

    @classmethod
    def from_json(cls, json_dict: Dict) -> Facet:
        return cls(json_dict["FieldName"])

    def to_json(self) -> Dict:
        super_json = super().to_json()
        super_json.update({"FieldName": self.field_name})
        return super_json


class FacetSetup:
    def __init__(self, facets: List[Facet] = None, range_facets: List[RangeFacet] = None):
        self.facets = facets or []
        self.range_facets = range_facets or []
        self.Id: Optional[str] = None

    @property
    def id_(self):
        return self.Id

    @id_.setter
    def id_(self, value: str):
        self.Id = value

    @classmethod
    def from_json(cls, json_dict: Dict) -> FacetSetup:
        return cls(
            [Facet.from_json(facet_json) for facet_json in json_dict["Facets"]],
            [RangeFacet.from_json(range_facet_json) for range_facet_json in json_dict["RangeFacets"]],
        )

    def to_json(self) -> Dict:
        return {
            "Facets": [facet.to_json() for facet in self.facets],
            "RangeFacets": [range_facet.to_json() for range_facet in self.range_facets],
        }



from __future__ import annotations
from enum import Enum
from typing import Union, List, Dict


class SpatialFieldType(Enum):
    GEOGRAPHY = "Geography"
    CARTESIAN = "Cartesian"

    def __str__(self):
        return self.value


class SpatialRelation(Enum):
    WITHIN = "Within"
    CONTAINS = "Contains"
    DISJOINT = "Disjoint"
    INTERSECTS = "Intersects"


class SpatialSearchStrategy(Enum):
    GEOHASH_PREFIX_TREE = "GeohashPrefixTree"
    QUAD_PREFIX_TREE = "QuadPrefixTree"
    BOUNDING_BOX = "BoundingBox"

    def __str__(self):
        return self.value


class SpatialUnits:
    KILOMETERS = "Kilometers"
    MILES = "Miles"


class SpatialOptions:
    DEFAULT_GEOHASH_LEVEL = 9
    DEFAULT_QUAD_TREE_LEVEL = 23

    def __init__(
        self,
        field_type: SpatialFieldType = SpatialFieldType.GEOGRAPHY,
        strategy: SpatialSearchStrategy = SpatialSearchStrategy.GEOHASH_PREFIX_TREE,
        max_tree_level: int = DEFAULT_GEOHASH_LEVEL,
        min_x: int = -180,
        max_x: int = 180,
        min_y: int = -90,
        max_y: int = 90,
        units: SpatialUnits = SpatialUnits.KILOMETERS,
    ):
        self.type = field_type
        self.strategy = strategy
        self.max_tree_level = max_tree_level
        self.min_x = min_x
        self.max_x = max_x
        self.min_y = min_y
        self.max_y = max_y
        self.units = units

    def to_json(self) -> Dict:
        return {
            "Type": self.type.value,
            "Strategy": self.strategy.value,
            "MaxTreeLevel": self.max_tree_level,
            "MinX": self.min_x,
            "MaxX": self.max_x,
            "MinY": self.min_y,
            "MaxY": self.max_y,
            "Units": self.units,
        }

    @classmethod
    def from_copy(cls, options: SpatialOptions) -> SpatialOptions:
        return cls(
            options.type,
            options.strategy,
            options.max_tree_level,
            options.min_x,
            options.max_x,
            options.min_y,
            options.max_y,
            options.units,
        )

    def __eq__(self, other):
        if self == other:
            return True
        if other is None:
            return False
        if type(self) != type(other):
            return False

        other: SpatialOptions

        result = self.type == other.type and self.strategy == other.strategy
        if self.type == SpatialFieldType.GEOGRAPHY:
            result = result and self.units == other.units
        if self.strategy != SpatialSearchStrategy.BOUNDING_BOX:
            result = result and self.max_tree_level == other.max_tree_level
            if self.type == SpatialFieldType.CARTESIAN:
                result = (
                    result
                    and self.min_x == other.min_x
                    and self.max_x == other.max_x
                    and self.min_y == other.min_y
                    and self.max_y == other.max_y
                )
        return result

    # todo: hashCode() override


class AutoSpatialMethodType(Enum):
    POINT = "Point"
    WKT = "Wkt"

    def __str__(self):
        return self.value


class AutoSpatialOptions(SpatialOptions):
    def __init__(
        self,
        field_type: SpatialFieldType = SpatialFieldType.GEOGRAPHY,
        strategy: SpatialSearchStrategy = SpatialSearchStrategy.GEOHASH_PREFIX_TREE,
        max_tree_level: int = SpatialOptions.DEFAULT_GEOHASH_LEVEL,
        min_x: int = -180,
        max_x: int = 180,
        min_y: int = -90,
        max_y: int = 90,
        units: SpatialUnits = SpatialUnits.KILOMETERS,
    ):
        super(AutoSpatialOptions, self).__init__(
            field_type, strategy, max_tree_level, min_x, max_x, min_y, max_y, units
        )
        self.method_type: Union[None, AutoSpatialMethodType] = None
        self.method_argument: Union[None, List[str]] = None

    @classmethod
    def from_json(cls, json_dict: Dict) -> AutoSpatialOptions:
        return cls.from_copy(SpatialOptions.from_copy(json_dict.get("Options")))


class SpatialOptionsFactory:
    class GeographySpatialOptionsFactory:
        def default_option(self, circle_radius_units: SpatialUnits = SpatialUnits.KILOMETERS) -> SpatialOptions:
            return self.geohash_prefix_tree_index(0, circle_radius_units)

        def bounding_box_index(self, circle_radius_units: SpatialUnits = SpatialUnits.KILOMETERS) -> SpatialOptions:
            ops = SpatialOptions()
            ops.type = SpatialFieldType.GEOGRAPHY
            ops.strategy = SpatialSearchStrategy.BOUNDING_BOX
            ops.units = circle_radius_units
            return ops

        def geohash_prefix_tree_index(
            self, max_tree_level: int, circle_radius_units: SpatialUnits = SpatialUnits.KILOMETERS
        ) -> SpatialOptions:
            if max_tree_level == 0:
                max_tree_level = SpatialOptions.DEFAULT_GEOHASH_LEVEL

            opts = SpatialOptions()
            opts.type = SpatialFieldType.GEOGRAPHY
            opts.max_tree_level = max_tree_level
            opts.strategy = SpatialSearchStrategy.GEOHASH_PREFIX_TREE
            opts.units = circle_radius_units
            return opts

        def quad_prefix_tree_index(
            self, max_tree_level: int, circle_radius_units: SpatialUnits = SpatialUnits.KILOMETERS
        ) -> SpatialOptions:
            if max_tree_level == 0:
                max_tree_level = SpatialOptions.DEFAULT_QUAD_TREE_LEVEL

            opts = SpatialOptions()
            opts.type = SpatialFieldType.GEOGRAPHY
            opts.max_tree_level = max_tree_level
            opts.strategy = SpatialSearchStrategy.QUAD_PREFIX_TREE
            opts.units = circle_radius_units
            return opts

    class CartesianSpatialOptionsFactory:
        def bounding_box_index(self) -> SpatialOptions:
            opts = SpatialOptions()
            opts.type = SpatialFieldType.CARTESIAN
            opts.strategy = SpatialSearchStrategy.BOUNDING_BOX
            return opts

        def quad_prefix_tree_index(
            self, max_tree_level: int, bounds: SpatialOptionsFactory.SpatialBounds
        ) -> SpatialOptions:
            if max_tree_level == 0:
                raise ValueError("maxTreeLevel")

            opts = SpatialOptions()
            opts.type = SpatialFieldType.CARTESIAN
            opts.max_tree_level = max_tree_level
            opts.strategy = SpatialSearchStrategy.QUAD_PREFIX_TREE
            opts.min_x = bounds.min_x
            opts.max_x = bounds.max_x
            opts.min_y = bounds.min_y
            opts.max_y = bounds.max_y

            return opts

    def geography(self) -> GeographySpatialOptionsFactory:
        return SpatialOptionsFactory.GeographySpatialOptionsFactory()

    def cartesian(self) -> CartesianSpatialOptionsFactory:
        return SpatialOptionsFactory.CartesianSpatialOptionsFactory()

    class SpatialBounds:
        def __init__(self, min_x: float, min_y: float, max_x: float, max_y: float):
            self.min_x = min_x
            self.min_y = min_y
            self.max_x = max_x
            self.max_y = max_y

        def __hash__(self):
            hash((self.max_x, self.min_x, self.max_y, self.min_y))

        def __eq__(self, other: object):
            if self == other:
                return True
            if other == None:
                return False
            if self.__class__ != other.__class__:
                return False
            other: SpatialOptionsFactory.SpatialBounds
            return all(
                [
                    self.max_x == other.max_x,
                    self.max_y == other.max_y,
                    self.min_x == other.min_x,
                    self.min_y == other.min_y,
                ]
            )



from typing import Optional, Any, Dict


class AnalyzerDefinition:
    def __init__(self, name: Optional[str] = None, code: Optional[str] = None):
        self.name = name
        self.code = code

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> "AnalyzerDefinition":
        return cls(json_dict["Name"], json_dict["Code"])

    def to_json(self) -> Dict[str, Any]:
        return {"Name": self.name, "Code": self.code}



from __future__ import annotations
import threading
from typing import Generic, Callable, Any, Union, TypeVar

T = TypeVar("T")


class Lazy(Generic[T]):
    def __init__(self, value_factory: Callable[[], Any]):
        self.__value_created = False  # todo: check if it isn't cached - volatile in java
        self.__value_factory = value_factory
        self.__value: Union[None, T] = None
        self.__value_lock = threading.Lock()

    @property
    def is_value_created(self) -> bool:
        return self.__value_created

    @property
    def value(self) -> T:  # double check locking
        if self.__value_created:
            return self.__value
        with self.__value_lock:
            if not self.__value_created:
                self.__value = self.__value_factory()
                self.__value_created = True
        return self.__value



from enum import Enum


class WhereOperator(Enum):
    EQUALS = "Equals"
    NOT_EQUALS = "NotEquals"
    GREATER_THAN = "GreaterThan"
    GREATER_THAN_OR_EQUAL = "GreaterThanOrEqual"
    LESS_THAN = "LessThan"
    LESS_THAN_OR_EQUAL = "LessThanOrEqual"
    IN = "In"
    ALL_IN = "AllIn"
    BETWEEN = "Between"
    SEARCH = "Search"
    LUCENE = "Lucene"
    STARTS_WITH = "StartsWith"
    ENDS_WITH = "EndsWith"
    EXISTS = "Exists"
    SPATIAL_WITHIN = "SpatialWithin"
    SPATIAL_CONTAINS = "SpatialContains"
    SPATIAL_DISJOINT = "SpatialDisjoints"
    SPATIAL_INTERSECTS = "SpatialIntersects"
    REGEX = "Regex"
    VECTOR_SEARCH = "VectorSearch"



from __future__ import annotations
from typing import Any, Dict

from ravendb.primitives import constants
from ravendb.documents.session.concurrency_check_mode import ConcurrencyCheckMode
from ravendb.json.metadata_as_dictionary import MetadataAsDictionary


class DocumentInfo:
    def __init__(
        self,
        key: str = None,
        change_vector: str = None,
        concurrency_check_mode: ConcurrencyCheckMode = None,
        ignore_changes: bool = None,
        metadata: dict = None,
        document: dict = None,
        metadata_instance: MetadataAsDictionary = None,
        entity: Any = None,
        new_document: bool = None,
        collection: str = None,
    ):
        self.key = key
        self.change_vector = change_vector
        self.concurrency_check_mode = concurrency_check_mode
        self.ignore_changes = ignore_changes
        self.metadata = metadata
        self.document = document
        self.metadata_instance = metadata_instance
        self.entity = entity
        self.new_document = new_document
        self.collection = collection

    @classmethod
    def get_new_document_info(cls, document: Dict) -> DocumentInfo:
        metadata = document.get(constants.Documents.Metadata.KEY)
        if not metadata:
            raise ValueError("Document must have a metadata")
        key = metadata.get(constants.Documents.Metadata.ID)
        if not key or not isinstance(key, str):
            raise ValueError("Document must have a key")

        change_vector = metadata.get(constants.Documents.Metadata.CHANGE_VECTOR)
        if not change_vector or not isinstance(change_vector, str):
            raise ValueError(f"Document {key} must have a Change Vector")

        return cls(key=key, document=document, metadata=metadata, entity=None, change_vector=change_vector)



import datetime
import json
from enum import Enum
from typing import TYPE_CHECKING, Type, TypeVar, List, Optional, Dict

from ravendb.documents.commands.multi_get import GetRequest, GetResponse
from ravendb.documents.operations.lazy.definition import LazyOperation
from ravendb.documents.session.operations.operations import GetRevisionOperation
from ravendb.json.result import JsonArrayResult
from ravendb.json.metadata_as_dictionary import MetadataAsDictionary

if TYPE_CHECKING:
    from ravendb.documents.session.document_session import DocumentSession
    from ravendb.documents.queries.query import QueryResult
    from ravendb.documents.store.lazy import Lazy


_T = TypeVar("_T")


class LazyRevisionOperations:
    def __init__(self, delegate: "DocumentSession"):
        self._delegate = delegate

    def get_by_change_vector(self, change_vector: str, object_type: Type[_T]) -> "Lazy[_T]":
        operation = GetRevisionOperation.from_change_vector(self._delegate, change_vector)
        lazy_revision_operation = LazyRevisionOperation(operation, LazyRevisionOperation.Mode.SINGLE, object_type)
        return self._delegate.add_lazy_operation(object_type, lazy_revision_operation, None)

    def get_metadata_for(
        self, id_: str, start: Optional[int] = 0, page_size: Optional[int] = 25
    ) -> "Lazy[List[MetadataAsDictionary]]":
        operation = GetRevisionOperation.from_start_page(self._delegate, id_, start, page_size)
        lazy_revision_operation = LazyRevisionOperation(
            operation, LazyRevisionOperation.Mode.LIST_OF_METADATA, MetadataAsDictionary
        )
        return self._delegate.add_lazy_operation(list, lazy_revision_operation, None)

    def get_by_change_vectors(
        self, change_vectors: List[str], object_type: Optional[Type[_T]] = None
    ) -> "Lazy[Dict[str, _T]]":
        operation = GetRevisionOperation.from_change_vectors(self._delegate, change_vectors)
        lazy_revision_operation = LazyRevisionOperation(operation, LazyRevisionOperation.Mode.MAP, object_type)
        return self._delegate.add_lazy_operation(dict, lazy_revision_operation, None)

    def get_by_before_date(
        self, id_: str, date: datetime.datetime, object_type: Optional[Type[_T]] = None
    ) -> "Lazy[_T]":
        operation = GetRevisionOperation.from_before_date(self._delegate, id_, date)
        lazy_revision_operation = LazyRevisionOperation(operation, LazyRevisionOperation.Mode.SINGLE, object_type)
        return self._delegate.add_lazy_operation(object_type, lazy_revision_operation, None)

    def get_for(
        self, id_: str, object_type: Optional[Type[_T]] = None, start: Optional[int] = 0, page_size: Optional[int] = 25
    ) -> "Lazy[List[_T]]":
        operation = GetRevisionOperation.from_start_page(self._delegate, id_, start, page_size)
        lazy_revision_operation = LazyRevisionOperation(operation, LazyRevisionOperation.Mode.MULTI, object_type)
        return self._delegate.add_lazy_operation(list, lazy_revision_operation, None)


class LazyRevisionOperation(LazyOperation):
    class Mode(Enum):
        SINGLE = "Single"
        MULTI = "Multi"
        MAP = "Map"
        LIST_OF_METADATA = "ListOfMetadata"

    def __init__(
        self, get_revision_operation: GetRevisionOperation, mode: Mode, object_type: Optional[Type[_T]] = None
    ):
        self.result: Optional[object] = None
        self.query_result: Optional[QueryResult] = None
        self._requires_retry: Optional[bool] = None

        self._object_type = object_type
        self._get_revision_operation = get_revision_operation
        self._mode = mode

    @property
    def requires_retry(self) -> Optional[bool]:
        return self._requires_retry

    @requires_retry.setter
    def requires_retry(self, value: bool):
        self._requires_retry = value

    def create_request(self) -> Optional[GetRequest]:
        get_revisions_command = self._get_revision_operation.command
        sb = ["?"]
        get_revisions_command.get_request_query_string(sb)
        get_request = GetRequest()
        get_request.method = "GET"
        get_request.url = "/revisions"
        get_request.query = "".join(sb)
        return get_request

    def handle_response(self, response: GetResponse) -> None:
        if response.result is None:
            return
        response_as_dict = json.loads(response.result)

        if response_as_dict is None:
            return
        json_array = response_as_dict.get("Results")

        json_array_result = JsonArrayResult()
        json_array_result.results = json_array
        self._get_revision_operation.set_result(json_array_result)

        if self._mode == LazyRevisionOperation.Mode.SINGLE:
            self.result = self._get_revision_operation.get_revision(self._object_type)
        elif self._mode == LazyRevisionOperation.Mode.MULTI:
            self.result = self._get_revision_operation.get_revisions_for(self._object_type)
        elif self._mode == LazyRevisionOperation.Mode.MAP:
            self.result = self._get_revision_operation.get_revisions(self._object_type)
        elif self._mode == LazyRevisionOperation.Mode.LIST_OF_METADATA:
            self.result = self._get_revision_operation.get_revisions_metadata_for()
        else:
            raise ValueError(f"Invalid mode: {self._mode}")



from __future__ import annotations

import datetime
import inspect
import json
import logging
import os
from concurrent.futures import ThreadPoolExecutor, Future, FIRST_COMPLETED, wait, ALL_COMPLETED
import uuid
from threading import Timer, Semaphore, Lock

import requests
from copy import copy

from ravendb.primitives import constants
from ravendb.documents.session.event_args import BeforeRequestEventArgs, FailedRequestEventArgs, SucceedRequestEventArgs
from ravendb.exceptions.exceptions import (
    AllTopologyNodesDownException,
    UnsuccessfulRequestException,
    DatabaseDoesNotExistException,
    AuthorizationException,
    RequestedNodeUnavailableException,
)
from ravendb.documents.operations.configuration.operations import GetClientConfigurationOperation
from ravendb.exceptions.exception_dispatcher import ExceptionDispatcher
from ravendb.exceptions.raven_exceptions import ClientVersionMismatchException


from ravendb.http.http_cache import HttpCache
from ravendb.http.misc import ReadBalanceBehavior, ResponseDisposeHandling, LoadBalanceBehavior, Broadcast
from ravendb.http.raven_command import RavenCommand, RavenCommandResponseType
from ravendb.http.server_node import ServerNode
from ravendb.http.topology import Topology, NodeStatus, NodeSelector, CurrentIndexAndNode, UpdateTopologyParameters
from ravendb.serverwide.commands import GetDatabaseTopologyCommand, GetClusterTopologyCommand

from http import HTTPStatus


from typing import TYPE_CHECKING, List, Dict, Tuple

if TYPE_CHECKING:
    from ravendb.documents.session import SessionInfo
    from ravendb.documents.conventions import DocumentConventions
    from typing import Union, Callable, Any, Optional


class RequestExecutor:
    __INITIAL_TOPOLOGY_ETAG = -2
    __GLOBAL_APPLICATION_IDENTIFIER = uuid.uuid4()
    CLIENT_VERSION = "7.0.0"
    logger = logging.getLogger("request_executor")

    # todo: initializer should take also cryptography certificates
    def __init__(
        self,
        database_name: str,
        conventions: DocumentConventions,
        certificate_path: Optional[str] = None,
        trust_store_path: Optional[str] = None,
        thread_pool_executor: Optional[ThreadPoolExecutor] = None,
        initial_urls: Optional[List[str]] = None,
    ):
        self.__update_topology_timer: Union[None, Timer] = None
        self.conventions = copy(conventions)
        self._node_selector: NodeSelector = None
        self.__default_timeout: datetime.timedelta = conventions.request_timeout
        self._cache: HttpCache = HttpCache()

        self.__certificate_path = certificate_path
        self.__trust_store_path = trust_store_path

        self.__topology_taken_from_node: Union[None, ServerNode] = None
        self.__update_client_configuration_semaphore = Semaphore(1)
        self.__update_database_topology_semaphore = Semaphore(1)

        self.__failed_nodes_timers: Dict[ServerNode, NodeStatus] = {}

        self._database_name = database_name

        self._last_returned_response: Union[None, datetime.datetime] = None

        self._thread_pool_executor = (
            ThreadPoolExecutor(max_workers=10) if not thread_pool_executor else thread_pool_executor
        )

        self.number_of_server_requests = 0

        self._topology_etag: Union[None, int] = None
        self._client_configuration_etag: Union[None, int] = None
        self._disable_topology_updates: Union[None, bool] = None
        self._disable_client_configuration_updates: Union[None, bool] = None
        self._last_server_version: Union[None, str] = None

        self.__http_session: Union[None, requests.Session] = None

        self.first_broadcast_attempt_timeout: Union[None, datetime.timedelta] = (
            conventions.first_broadcast_attempt_timeout
        )
        self.second_broadcast_attempt_timeout: Union[None, datetime.timedelta] = (
            conventions.second_broadcast_attempt_timeout
        )

        self._first_topology_update_task: Union[None, Future] = None
        self._last_known_urls: Union[None, List[str]] = None
        self._disposed: Union[None, bool] = None

        self.__synchronized_lock = Lock()

        # --- events ---
        self._on_before_request: List[Callable[[BeforeRequestEventArgs], Any]] = []
        self.__on_failed_request: List[Callable[[FailedRequestEventArgs], None]] = []
        self.__on_succeed_request: List[Callable[[SucceedRequestEventArgs], None]] = []
        self._on_topology_updated: List[Callable[[Topology], None]] = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if self._disposed:
            return

        self._disposed = True
        self._cache.close()

        if self.__update_topology_timer:
            self.__update_topology_timer.cancel()

        self._dispose_all_failed_nodes_timers()
        if self.__http_session is not None:
            self.__http_session.close()

    @property
    def certificate_path(self) -> str:
        return self.__certificate_path

    @property
    def trust_store_path(self) -> str:
        return self.__trust_store_path

    @property
    def url(self) -> Union[None, str]:
        if self._node_selector is None:
            return None

        preferred_node = self._node_selector.get_preferred_node()

        return preferred_node.current_node.url if preferred_node is not None else None

    @property
    def last_server_version(self):
        return self._last_server_version

    @property
    def topology_etag(self) -> int:
        return self._topology_etag

    @property
    def topology(self) -> Topology:
        return self._node_selector.topology if self._node_selector else None

    @property
    def http_session(self):
        http_session = self.__http_session
        if http_session:
            return http_session
        self.__http_session = self.__create_http_session()
        return self.__http_session

    def __create_http_session(self) -> requests.Session:
        session = requests.session()
        session.cert = self.__certificate_path
        session.verify = self.__trust_store_path if self.__trust_store_path else True
        return session

    @property
    def cache(self) -> HttpCache:
        return self._cache

    @property
    def topology_nodes(self) -> List[ServerNode]:
        return self.topology.nodes if self.topology else None

    @property
    def client_configuration_etag(self) -> int:
        return self._client_configuration_etag

    @property
    def default_timeout(self) -> datetime.timedelta:
        return self.__default_timeout

    @default_timeout.setter
    def default_timeout(self, value: datetime.timedelta) -> None:
        self.__default_timeout = value

    @property
    def preferred_node(self) -> CurrentIndexAndNode:
        self._ensure_node_selector()
        return self._node_selector.get_preferred_node()

    def get_requested_node(self, node_tag: str, throw_if_contains_failures: bool = False) -> CurrentIndexAndNode:
        self._ensure_node_selector()
        current_index_and_node = self._node_selector.get_requested_node(node_tag)

        if throw_if_contains_failures and not self._node_selector.node_is_available(
            current_index_and_node.current_index
        ):
            raise RequestedNodeUnavailableException(
                f"Requested node {node_tag} currently unavailable, please try again later."
            )

        return current_index_and_node

    def get_fastest_node(self) -> CurrentIndexAndNode:
        self._ensure_node_selector()

        return self._node_selector.get_fastest_node()

    def get_node_by_session_id(self, session_id: int = None) -> CurrentIndexAndNode:
        self._ensure_node_selector()

        return self._node_selector.get_node_by_session_id(session_id)

    def __on_failed_request_invoke(self, url: str, e: Exception):
        for event in self.__on_failed_request:
            event(FailedRequestEventArgs(self._database_name, url, e, None, None))

    def __on_failed_request_invoke_details(
        self, url: str, e: Exception, request: Optional[requests.Request], response: Optional[requests.Response]
    ) -> None:
        for event in self.__on_failed_request:
            event(FailedRequestEventArgs(self._database_name, url, e, request, response))

    def _on_succeed_request_invoke(
        self, database: str, url: str, response: requests.Response, request: requests.Request, attempt_number: int
    ):
        for event in self.__on_succeed_request:
            event(SucceedRequestEventArgs(database, url, response, request, attempt_number))

    def _on_topology_updated_invoke(self, topology: Topology) -> None:
        for event in self._on_topology_updated:
            event(topology)

    def add_on_succeed_request(self, event: Callable[[SucceedRequestEventArgs], None]):
        self.__on_succeed_request.append(event)

    def remove_on_succeed_request(self, event: Callable[[SucceedRequestEventArgs], None]):
        self.__on_succeed_request.remove(event)

    def add_on_failed_request(self, event: Callable[[FailedRequestEventArgs], None]):
        self.__on_failed_request.append(event)

    def remove_on_failed_request(self, event: Callable[[FailedRequestEventArgs], None]):
        self.__on_failed_request.remove(event)

    def add_on_before_request(self, event: Callable[[BeforeRequestEventArgs], None]):
        self._on_before_request.append(event)

    def remove_on_before_request(self, event: Callable[[BeforeRequestEventArgs], None]):
        self._on_before_request.remove(event)

    def add_on_topology_updated(self, event: Callable[[Topology], None]):
        self._on_topology_updated.append(event)

    def remove_on_topology_updated(self, event: Callable[[Topology], None]):
        self._on_topology_updated.remove(event)

    @classmethod
    def create(
        cls,
        initial_urls: List[str],
        database_name: str,
        conventions: DocumentConventions,
        certificate_path: Optional[str] = None,
        trust_store_path: Optional[str] = None,
        thread_pool_executor: Optional[ThreadPoolExecutor] = None,
    ) -> RequestExecutor:
        executor = cls(database_name, conventions, certificate_path, trust_store_path, thread_pool_executor)
        executor._first_topology_update_task = executor._first_topology_update(
            initial_urls, cls.__GLOBAL_APPLICATION_IDENTIFIER
        )
        return executor

    @classmethod
    def create_for_single_node_with_configuration_updates(
        cls,
        url: str,
        database_name: str,
        conventions: DocumentConventions,
        certificate_path: Optional[str] = None,
        trust_store_path: Optional[str] = None,
        thread_pool_executor: Optional[ThreadPoolExecutor] = None,
    ) -> RequestExecutor:
        executor = cls.create_for_single_node_without_configuration_updates(
            url, database_name, conventions, certificate_path, trust_store_path, thread_pool_executor
        )
        executor._disable_client_configuration_updates = False
        return executor

    @classmethod
    def create_for_single_node_without_configuration_updates(
        cls,
        url: str,
        database_name: str,
        conventions: DocumentConventions,
        certificate_path: Optional[str] = None,
        trust_store_path: Optional[str] = None,
        thread_pool_executor: Optional[ThreadPoolExecutor] = None,
    ) -> RequestExecutor:
        initial_urls = cls.validate_urls([url])
        executor = cls(database_name, conventions, certificate_path, trust_store_path, thread_pool_executor)

        topology = Topology(-1, [ServerNode(initial_urls[0], database_name)])
        executor._node_selector = NodeSelector(topology, thread_pool_executor)
        executor._topology_etag = cls.__INITIAL_TOPOLOGY_ETAG
        executor._disable_topology_updates = True
        executor._disable_client_configuration_updates = True

        return executor

    def _update_client_configuration_async(self, server_node: ServerNode) -> Future:
        if self._disposed:
            future = Future()
            future.set_result(None)
            return future

        def __run_async():
            self.__update_client_configuration_semaphore.acquire()

            old_disable_client_configuration_updates = self._disable_client_configuration_updates
            self._disable_client_configuration_updates = True

            try:
                if self._disposed:
                    return

                command = GetClientConfigurationOperation.GetClientConfigurationCommand()
                self.execute(server_node, None, command, False, None)

                result = command.result
                if not result:
                    return

                self.conventions.update_from(result.configuration)
                self.client_configuration_etag = result.etag
            finally:
                self._disable_client_configuration_updates = old_disable_client_configuration_updates
                self.__update_client_configuration_semaphore.release()

        return self._thread_pool_executor.submit(__run_async)

    def update_topology_async(self, parameters: UpdateTopologyParameters) -> Future:
        if not parameters:
            raise ValueError("Parameters cannot be None")

        if self._disable_topology_updates:
            fut = Future()
            fut.set_result(False)
            return fut

        if self._disposed:
            fut = Future()
            fut.set_result(False)
            return fut

        def __supply_async():
            # prevent double topology updates if execution takes too much time
            # --> in cases with transient issues
            lock_taken = self.__update_database_topology_semaphore.acquire(timeout=parameters.timeout_in_ms)
            if not lock_taken:
                return False

            try:
                if self._disposed:
                    return False

                command = GetDatabaseTopologyCommand(
                    parameters.debug_tag,
                    parameters.application_identifier if self.conventions.send_application_identifier else None,
                )
                self.execute(parameters.node, None, command, False, None)
                topology = command.result

                if self._node_selector is None:
                    self._node_selector = NodeSelector(topology, self._thread_pool_executor)

                    if self.conventions.read_balance_behavior == ReadBalanceBehavior.FASTEST_NODE:
                        self._node_selector.schedule_speed_test()

                elif self._node_selector.on_update_topology(topology, parameters.force_update):
                    self._dispose_all_failed_nodes_timers()
                    if self.conventions.read_balance_behavior == ReadBalanceBehavior.FASTEST_NODE:
                        self._node_selector.schedule_speed_test()

                self._topology_etag = self._node_selector.topology.etag

                self._on_topology_updated_invoke(topology)
            except Exception as e:
                if not self._disposed:
                    raise e

            finally:
                self.__update_database_topology_semaphore.release()
            return True

        return self._thread_pool_executor.submit(__supply_async)

    def _first_topology_update(self, input_urls: List[str], application_identifier: Union[None, uuid.UUID]) -> Future:
        initial_urls = self.validate_urls(input_urls)
        errors: List[Tuple[str, Exception]] = []

        def __run(errors: list):
            for url in initial_urls:
                try:
                    server_node = ServerNode(url, self._database_name)
                    update_parameters = UpdateTopologyParameters(server_node)
                    update_parameters.timeout_in_ms = 0x7FFFFFFF
                    update_parameters.debug_tag = "first-topology-update"
                    update_parameters.application_identifier = application_identifier

                    self.update_topology_async(update_parameters).result()
                    self.__initialize_update_topology_timer()

                    self.__topology_taken_from_node = server_node
                    return
                except Exception as e:
                    if isinstance(e.__cause__, AuthorizationException):
                        self._last_known_urls = initial_urls
                        raise e.__cause__

                    if isinstance(e.__cause__, DatabaseDoesNotExistException):
                        self._last_known_urls = initial_urls
                        raise e.__cause__

                    errors.append((url, e))

            topology = Topology(
                self._topology_etag,
                (
                    self.topology_nodes
                    if self.topology_nodes
                    else list(map(lambda url_val: ServerNode(url_val, self._database_name, "!"), initial_urls))
                ),
            )

            self._node_selector = NodeSelector(topology, self._thread_pool_executor)

            if initial_urls:
                self.__initialize_update_topology_timer()
                return

            self._last_known_urls = initial_urls
            # todo: details from exceptions in inner_list

        return self._thread_pool_executor.submit(__run, errors)

    @staticmethod
    def validate_urls(initial_urls: List[str]) -> List[str]:
        # todo: implement validation
        return initial_urls

    def __initialize_update_topology_timer(self):
        if self.__update_topology_timer is not None:
            return

        with self.__synchronized_lock:
            if self.__update_topology_timer is not None:
                return

            self.__update_topology_timer = Timer(60, self.__update_topology_callback)

    def _dispose_all_failed_nodes_timers(self) -> None:
        for node, status in self.__failed_nodes_timers.items():
            status.close()
        self.__failed_nodes_timers.clear()

    def execute_command(self, command: RavenCommand, session_info: Optional[SessionInfo] = None) -> None:
        topology_update = self._first_topology_update_task
        if (
            topology_update is not None
            and (topology_update.done() and (not topology_update.exception()) and (not topology_update.cancelled()))
            or self._disable_topology_updates
        ):
            current_index_and_node = self.choose_node_for_request(command, session_info)
            self.execute(
                current_index_and_node.current_node, current_index_and_node.current_index, command, True, session_info
            )
        else:
            self.__unlikely_execute(command, topology_update, session_info)

    def execute(
        self,
        chosen_node: ServerNode = None,
        node_index: int = None,
        command: RavenCommand = None,
        should_retry: bool = None,
        session_info: SessionInfo = None,
        ret_request: bool = False,
    ) -> Union[None, requests.Request]:
        if command.failover_topology_etag == RequestExecutor.__INITIAL_TOPOLOGY_ETAG:
            command.failover_topology_etag = RequestExecutor.__INITIAL_TOPOLOGY_ETAG
            if self._node_selector and self._node_selector.topology:
                topology = self._node_selector.topology
                if topology.etag:
                    command.failover_topology_etag = topology.etag

        request = self.__create_request(chosen_node, command)
        url = request.url

        if not request:
            return

        if ret_request:
            request_ref = request

        if not request:
            return

        no_caching = session_info.no_caching if session_info else False

        cached_item, change_vector, cached_value = self._get_from_cache(command, not no_caching, url)
        # todo: if change_vector exists try get from cache - aggressive caching
        with cached_item:
            # todo: try get from cache
            self._set_request_headers(session_info, change_vector, request)

            command.number_of_attempts = command.number_of_attempts + 1
            attempt_num = command.number_of_attempts
            for func in self._on_before_request:
                func(BeforeRequestEventArgs(self._database_name, url, request, attempt_num))
            response = self._send_request_to_server(
                chosen_node, node_index, command, should_retry, session_info, request, url
            )

            if response is None:
                return

            refresh_tasks = self._refresh_if_needed(chosen_node, response)

            command.status_code = response.status_code
            response_dispose = ResponseDisposeHandling.AUTOMATIC

            try:
                if response.status_code == HTTPStatus.NOT_MODIFIED:
                    self._on_succeed_request_invoke(self._database_name, url, response, request, attempt_num)
                    cached_item.not_modified()
                    if command.response_type == RavenCommandResponseType.OBJECT:
                        command.set_response(cached_value, True)
                    return

                if response.status_code >= 400:
                    if not self._handle_unsuccessful_response(
                        chosen_node,
                        node_index,
                        command,
                        request,
                        response,
                        url,
                        session_info,
                        should_retry,
                    ):
                        db_missing_header = response.headers.get("Database-Missing", None)
                        if db_missing_header is not None:
                            raise DatabaseDoesNotExistException(db_missing_header)
                        self._throw_failed_to_contact_all_nodes(command, request)
                    return  # we either handled this already in the unsuccessful response or we are throwing
                self._on_succeed_request_invoke(self._database_name, url, response, request, attempt_num)
                response_dispose = command.process_response(self._cache, response, url)
                self._last_returned_response = datetime.datetime.utcnow()
            finally:
                if response_dispose == ResponseDisposeHandling.AUTOMATIC:
                    response.close()
                if len(refresh_tasks) > 0:
                    try:
                        wait(refresh_tasks, return_when=ALL_COMPLETED)
                    except:
                        raise

    def _refresh_if_needed(self, chosen_node: ServerNode, response: requests.Response) -> List[Future]:
        refresh_topology = response.headers.get(constants.Headers.REFRESH_TOPOLOGY, False)
        refresh_client_configuration = response.headers.get(constants.Headers.REFRESH_CLIENT_CONFIGURATION, False)

        refresh_task = Future()
        refresh_task.set_result(False)

        refresh_client_configuration_task = Future()
        refresh_client_configuration_task.set_result(None)

        if refresh_topology:
            update_parameters = UpdateTopologyParameters(chosen_node)
            update_parameters.timeout_in_ms = 0
            update_parameters.debug_tag = "refresh-topology-header"
            refresh_task = self.update_topology_async(update_parameters)

        if refresh_client_configuration:
            refresh_client_configuration_task = self._update_client_configuration_async(chosen_node)

        return [refresh_task, refresh_client_configuration_task]

    def _send_request_to_server(
        self,
        chosen_node: ServerNode,
        node_index: int,
        command: RavenCommand,
        should_retry: bool,
        session_info: SessionInfo,
        request: requests.Request,
        url: str,
    ) -> Optional[requests.Response]:
        try:
            self.number_of_server_requests += 1
            timeout = command.timeout if command.timeout else self.__default_timeout

            if not timeout:
                return self.__send(chosen_node, command, session_info, request)

            else:
                try:
                    return self.__send(chosen_node, command, session_info, request)
                except requests.Timeout as t:
                    if not should_retry:
                        if command.failed_nodes is None:
                            command.failed_nodes = {}
                        command.failed_nodes[chosen_node] = t
                        raise t

                    if not self.__handle_server_down(
                        url, chosen_node, node_index, command, request, None, t, session_info, should_retry
                    ):
                        self._throw_failed_to_contact_all_nodes(command, request)

                    return None
        except IOError as e:
            if not should_retry:
                raise

            if not self.__handle_server_down(
                url, chosen_node, node_index, command, request, None, e, session_info, should_retry
            ):
                self._throw_failed_to_contact_all_nodes(command, request)

            return None

    def __send(
        self, chosen_node: ServerNode, command: RavenCommand, session_info: SessionInfo, request: requests.Request
    ) -> requests.Response:
        response: Optional[requests.Response] = None

        if self.should_execute_on_all(chosen_node, command):
            response = self.__execute_on_all_to_figure_out_the_fastest(chosen_node, command)
        else:
            response = command.send(self.http_session, request)

        # PERF: The reason to avoid rechecking every time is that servers wont change so rapidly
        #       and therefore we dismish its cost by orders of magnitude just doing it
        #       once in a while. We dont care also about the potential race conditions that may happen
        #       here mainly because the idea is to have a lax mechanism to recheck that is at least
        #       orders of magnitude faster than currently.
        if chosen_node.should_update_server_version():
            server_version = self.__try_get_server_version(response)
            if server_version is not None:
                chosen_node.update_server_version(server_version)

        self._last_server_version = chosen_node.last_server_version

        if session_info and session_info.last_cluster_transaction_index:
            # if we reach here it means that sometime a cluster transaction has occurred against this database.
            # Since the current executed command can be dependent on that, we have to wait for the cluster transaction.
            # But we can't do that if the server is an old one.

            if not self._last_server_version or self._last_server_version.lower() < "4.1":
                raise ClientVersionMismatchException(
                    f"The server on {chosen_node.url} has an old version and"
                    f" can't perform the command since this command dependent on"
                    f" a cluster transaction which this node doesn't support"
                )

        return response

    def choose_node_for_request(self, cmd: RavenCommand, session_info: SessionInfo) -> CurrentIndexAndNode:
        # When we disable topology updates we cannot rely on the node tag,
        # Because the initial topology will not have them
        if not self._disable_topology_updates:
            if cmd.selected_node_tag and not cmd.selected_node_tag.isspace():
                return self._node_selector.get_requested_node(cmd.selected_node_tag)

        if self.conventions.load_balance_behavior == LoadBalanceBehavior.USE_SESSION_CONTEXT:
            if session_info is not None and session_info.can_use_load_balance_behavior:
                return self._node_selector.get_node_by_session_id(session_info.session_id)

        if not cmd.is_read_request:
            return self._node_selector.get_preferred_node()

        if self.conventions.read_balance_behavior == ReadBalanceBehavior.NONE:
            return self._node_selector.get_preferred_node()
        elif self.conventions.read_balance_behavior == ReadBalanceBehavior.ROUND_ROBIN:
            return self._node_selector.get_node_by_session_id(session_info.session_id if session_info else 0)
        elif self.conventions.read_balance_behavior == ReadBalanceBehavior.FASTEST_NODE:
            return self._node_selector.get_fastest_node()
        raise RuntimeError()

    def __unlikely_execute(
        self, command: RavenCommand, topology_update: Union[None, Future[None]], session_info: SessionInfo
    ) -> None:
        self.__wait_for_topology_update(topology_update)

        current_index_and_node = self.choose_node_for_request(command, session_info)
        self.execute(
            current_index_and_node.current_node, current_index_and_node.current_index, command, True, session_info
        )

    def __wait_for_topology_update(self, topology_update: Future[None]) -> None:
        try:
            if topology_update is None or topology_update.exception():
                with self.__synchronized_lock:
                    if self._first_topology_update_task is None or topology_update == self._first_topology_update_task:
                        if self._last_known_urls is None:
                            # shouldn't happen
                            raise RuntimeError(
                                "No known topology and no previously known one, cannot proceed, likely a bug"
                            )
                        self._first_topology_update_task = self._first_topology_update(self._last_known_urls, None)
                    topology_update = self._first_topology_update_task
            topology_update.result()
        # todo: narrow the exception scope down
        except BaseException as e:
            with self.__synchronized_lock:
                if self._first_topology_update_task == topology_update:
                    self._first_topology_update_task = None  # next request will raise it

            raise e

    def __update_topology_callback(self) -> None:
        time = datetime.datetime.now()
        if (time - self._last_returned_response.time) <= datetime.timedelta(minutes=5):
            return

        try:
            selector = self._node_selector
            if selector is None:
                return
            preferred_node = selector.get_preferred_node()
            server_node = preferred_node.current_node
        except Exception as e:
            self.logger.info("Couldn't get preferred node Topology from _updateTopologyTimer", exc_info=e)
            return

        update_parameters = UpdateTopologyParameters(server_node)
        update_parameters.timeout_in_ms = 0
        update_parameters.debug_tag = "timer-callback"

        try:
            self.update_topology_async(update_parameters)
        except Exception as e:
            self.logger.info("Couldn't update topology from __update_topology_timer", exc_info=e)

    def _set_request_headers(
        self, session_info: SessionInfo, cached_change_vector: Union[None, str], request: requests.Request
    ) -> None:
        if cached_change_vector is not None:
            request.headers[constants.Headers.IF_NONE_MATCH] = f'"{cached_change_vector}"'

        if not self._disable_client_configuration_updates:
            request.headers[constants.Headers.CLIENT_CONFIGURATION_ETAG] = f'"{self._client_configuration_etag}"'

        if session_info and session_info.last_cluster_transaction_index:
            request.headers[constants.Headers.LAST_KNOWN_CLUSTER_TRANSACTION_INDEX] = str(
                session_info.last_cluster_transaction_index
            )

        if not self._disable_topology_updates:
            request.headers[constants.Headers.TOPOLOGY_ETAG] = f'"{self._topology_etag}"'

        if not request.headers.get(constants.Headers.CLIENT_VERSION):
            request.headers[constants.Headers.CLIENT_VERSION] = RequestExecutor.CLIENT_VERSION

    def _get_from_cache(
        self, command: RavenCommand, use_cache: bool, url: str
    ) -> Tuple[HttpCache.ReleaseCacheItem, Optional[str], Optional[str]]:
        if (
            use_cache
            and command.can_cache
            and command.is_read_request
            and command.response_type == RavenCommandResponseType.OBJECT
        ):
            return self._cache.get(url)

        return HttpCache.ReleaseCacheItem(), None, None

    @staticmethod
    def __try_get_server_version(response: requests.Response) -> Union[None, str]:
        server_version_header = response.headers.get(constants.Headers.SERVER_VERSION)

        if server_version_header is not None:
            return server_version_header

    def _throw_failed_to_contact_all_nodes(self, command: RavenCommand, request: requests.Request):
        if not command.failed_nodes:
            raise RuntimeError(
                "Received unsuccessful response and couldn't recover from it. "
                "Also, no record of exceptions per failed nodes. This is weird and should not happen."
            )

        if len(command.failed_nodes) == 1:
            raise command.failed_nodes.popitem()

        message = (
            f"Tried to send {command._result_class.__name__} request via {request.method}"
            f" {request.url} to all configured nodes in the topology, none of the attempt succeeded.{os.linesep}"
        )

        if self.__topology_taken_from_node is not None:
            message += (
                f"I was able to fetch {self.__topology_taken_from_node.database}"
                f" topology from {self.__topology_taken_from_node.url}.{os.linesep}"
            )

        nodes = None

        if self._node_selector and self._node_selector.topology:
            nodes = self._node_selector.topology.nodes

        if nodes is None:
            message += "Topology is empty."
        else:
            message += "Topology: "

            for node in nodes:
                exception = command.failed_nodes.get(node)
                message += (
                    f"{os.linesep}"
                    f"[Url: {node.url}, "
                    f"ClusterTag: {node.cluster_tag}, "
                    f"Exception: {exception.args[0] if exception else 'No exception'}]"
                )

        raise AllTopologyNodesDownException(message)

    def should_execute_on_all(self, chosen_node: ServerNode, command: RavenCommand) -> bool:
        return (
            self.conventions.read_balance_behavior == ReadBalanceBehavior.FASTEST_NODE
            and self._node_selector
            and self._node_selector.in_speed_test_phase
            and len(self._node_selector.topology.nodes) > 1
            and command.is_read_request()
            and command.response_type == RavenCommandResponseType.OBJECT
            and chosen_node is not None
            and not isinstance(command, Broadcast)
        )

    def __execute_on_all_to_figure_out_the_fastest(
        self, chosen_node: ServerNode, command: RavenCommand
    ) -> requests.Response:
        number_failed_tasks = [0]  # mutable integer
        preferred_task: Future[RequestExecutor.IndexAndResponse] = None

        nodes = self._node_selector.topology.nodes
        tasks: List[Future[RequestExecutor.IndexAndResponse]] = [None] * len(nodes)

        for i in range(len(nodes)):
            task_number = i
            self.number_of_server_requests += 1

            def __supply_async(
                single_element_list_number_failed_tasks: List[int],
            ) -> (RequestExecutor.IndexAndResponse, int):
                try:
                    request, str_ref = self.__create_request(nodes[task_number], command)
                    self._set_request_headers(None, None, request)
                    return self.IndexAndResponse(task_number, command.send(self.http_session, request))
                except Exception as e:
                    single_element_list_number_failed_tasks[0] += 1
                    tasks[task_number] = None
                    raise RuntimeError("Request execution failed", e)

            task = self._thread_pool_executor.submit(__supply_async, number_failed_tasks)

            if nodes[i].cluster_tag == chosen_node.cluster_tag:
                preferred_task = task
            else:
                task.add_done_callback(lambda result: result.response.close())

            tasks[i] = task

        while number_failed_tasks[0] < len(tasks):
            try:
                first_finished, slow = wait(list(filter(lambda t: t is not None, tasks)), return_when=FIRST_COMPLETED)
                fastest = first_finished.pop().result()
                fastest: RequestExecutor.IndexAndResponse
                self._node_selector.record_fastest(fastest.index, nodes[fastest.index])
                break
            except BaseException:
                for i in range(len(nodes)):
                    if tasks[i].exception() is not None:
                        number_failed_tasks[0] += 1
                        tasks[i] = None

        # we can reach here if number of failed task equal to the number of the nodes,
        # in which case we have nothing to do

        return preferred_task.result().response

    def __create_request(self, node: ServerNode, command: RavenCommand) -> Optional[requests.Request]:
        request = command.create_request(node)
        # todo: optimize that if - look for the way to make less ifs each time
        if request.data and not isinstance(request.data, str) and not inspect.isgenerator(request.data):
            request.data = json.dumps(request.data, default=self.conventions.json_default_method)

        # todo: 1117 - 1133
        return request or None

    def should_broadcast(self, command: RavenCommand) -> bool:
        if not isinstance(command, Broadcast):
            return False

        topology_nodes = self.topology_nodes
        if not topology_nodes:
            return False

        return True

    class BroadcastState:
        def __init__(self, command: RavenCommand, index: int, node: ServerNode, request: requests.Request):
            self.command = command
            self.index = index
            self.node = node
            self.request = request

    def __broadcast(self, command: RavenCommand, session_info: SessionInfo) -> object:
        if not isinstance(command, Broadcast):
            raise RuntimeError("You can broadcast only commands that inherit from Broadcast class.")
        command: Union[RavenCommand, Broadcast]
        failed_nodes = command.failed_nodes

        command.failed_nodes = {}
        broadcast_tasks: Dict[Future[None], RequestExecutor.BroadcastState] = {}

        try:
            self.__send_to_all_nodes(broadcast_tasks, session_info, command)

            return self.__wait_for_broadcast_result(command, broadcast_tasks)
        finally:
            for task, broadcast_state in broadcast_tasks.items():
                if task:

                    def __exceptionally(self_task: Future):
                        if self_task.exception():
                            index = broadcast_state.index
                            node = self._node_selector.topology.nodes[index]
                            if node in failed_nodes:
                                # if other node succeed in broadcast we need to send health checks to
                                # the original failed node
                                pass
                                # todo: health check
                        return None

                    task.add_done_callback(__exceptionally)

    def __wait_for_broadcast_result(self, command: RavenCommand, tasks: Dict[Future[None], BroadcastState]) -> object:
        while tasks:
            error = None
            try:
                wait(tasks.keys(), return_when=FIRST_COMPLETED)
            except Exception as e:
                error = e

            completed = filter(Future.done, tasks.keys()).__next__()

            if error is not None:
                failed = tasks.get(completed)
                node = self._node_selector.topology.nodes[failed.index]

                command.failed_nodes[node] = error

                self._node_selector.on_failed_request(failed.index)
                # todo: self.spawn_health_checks(node, failed.index)

                del tasks[completed]
                continue

            # no need to abort requests, 'slow' variable contains cancelled tasks due to completion of first task

            self._node_selector.restore_node_index(tasks.get(completed).index)
            return tasks.get(completed).command.result

        exceptions = ", ".join(
            list(map(lambda x: str(UnsuccessfulRequestException(x[0].url, x[1])), set(command.failed_nodes)))
        )

        raise AllTopologyNodesDownException(f"Broadcasting {command.__class__.__name__} failed: {exceptions}")

    def _handle_unsuccessful_response(
        self,
        chosen_node: ServerNode,
        node_index: int,
        command: RavenCommand,
        request: requests.Request,
        response: requests.Response,
        url: str,
        session_info: SessionInfo,
        should_retry: bool,
    ) -> bool:
        if response.status_code == HTTPStatus.NOT_FOUND:
            self._cache.set_not_found(url, False)  # todo : check if aggressively cached, don't just pass False
            if command.response_type == RavenCommandResponseType.EMPTY:
                return True
            elif command.response_type == RavenCommandResponseType.OBJECT:
                command.set_response(None, False)
            else:
                command.set_response_raw(response, None)
            return True

        elif response.status_code == HTTPStatus.FORBIDDEN:
            msg = self.__try_get_response_of_error(response)
            builder = ["Forbidden access to ", chosen_node.database, "@", chosen_node.url, ", "]
            if self.__certificate_path is None:
                builder.append("a certificate is required. ")
            else:
                builder.append("certificate does not have permission to access it or is unknown. ")

            builder.append(" Method: ")
            builder.append(request.method)
            builder.append(", Request: ")
            builder.append(request.url)
            builder.append(os.linesep)
            builder.append(msg)

            raise AuthorizationException("".join(builder))

        elif (
            response.status_code == HTTPStatus.GONE
        ):  # request not relevant for the chosen node - the database has been moved to a different one
            if not should_retry:
                return False

            if node_index is not None:
                self._node_selector.on_failed_request(node_index)

            if not command.failed_nodes is None:
                command.failed_nodes = {}

            if command.is_failed_with_node(chosen_node):
                command.failed_nodes[chosen_node] = UnsuccessfulRequestException(
                    f"Request to {request.url} ({request.method}) is not relevant for this node anymore."
                )

            index_and_node = self.choose_node_for_request(command, session_info)

            if index_and_node.current_node in command.failed_nodes:
                update_parameters = UpdateTopologyParameters(chosen_node)
                update_parameters.timeout_in_ms = 60000
                update_parameters.force_update = True
                update_parameters.debug_tag = "handle-unsuccessful-response"
                success = self.update_topology_async(update_parameters).result()
                if not success:
                    return False

                command.failed_nodes.clear()
                index_and_node = self.choose_node_for_request(command, session_info)
                self.execute(index_and_node.current_node, index_and_node.current_index, command, False, session_info)
                return True

        elif response.status_code in [
            HTTPStatus.GATEWAY_TIMEOUT,
            HTTPStatus.REQUEST_TIMEOUT,
            HTTPStatus.BAD_GATEWAY,
            HTTPStatus.SERVICE_UNAVAILABLE,
        ]:
            return self.__handle_server_down(
                url, chosen_node, node_index, command, request, response, None, session_info, should_retry
            )

        elif response.status_code == HTTPStatus.CONFLICT:
            data = json.loads(response.text)
            message = data.get("Message", None)
            err_type = data.get("Type", None)

            raise RuntimeError(f"{err_type}: {message}")  # todo: handle conflict (exception dispatcher involved)

        elif response.status_code == 425:  # too early
            if not should_retry:
                return False

            if node_index is not None:
                self._node_selector.on_failed_request(node_index)

            if command.failed_nodes is None:
                command.failed_nodes = {}

            if not command.is_failed_with_node(chosen_node):
                command.failed_nodes[chosen_node] = UnsuccessfulRequestException(
                    f"Request to '{request.url}' ({request.method}) is processing and not yed available on that node."
                )

            next_node = self.choose_node_for_request(command, session_info)
            self.execute(next_node.current_node, next_node.current_index, command, True, session_info)

            if node_index is not None:
                self._node_selector.restore_node_index(node_index)

            return True
        else:
            command.on_response_failure(response)
            raise RuntimeError(
                json.loads(response.text).get("Message", "Missing message")
            )  # todo: Exception dispatcher

        return False

    def __try_get_response_of_error(self, response: requests.Response) -> str:
        try:
            return response.content.decode("utf-8")
        except Exception as e:
            return f"Could not read request: {e.args[0]}"

    def __send_to_all_nodes(
        self, tasks: Dict[Future, BroadcastState], session_info: SessionInfo, command: Union[RavenCommand, Broadcast]
    ):
        for index in range(len(self._node_selector.topology.nodes)):
            state = self.BroadcastState(
                command.prepare_to_broadcast(self.conventions), index, self._node_selector.topology.nodes[index], None
            )

            state.command.timeout = self.second_broadcast_attempt_timeout

            # todo: aggressive caching options

            def __run_async() -> None:
                # todo: aggressive caching options
                try:
                    request = self.execute(state.node, None, state.command, False, session_info, True)
                    state.request = request
                finally:
                    # todo: aggressive_caching
                    pass

            task = self._thread_pool_executor.submit(__run_async)
            tasks[task] = state

    def __handle_server_down(
        self,
        url: str,
        chosen_node: ServerNode,
        node_index: Optional[int],
        command: RavenCommand,
        request: requests.Request,
        response: Optional[requests.Response],
        e: Exception,
        session_info: SessionInfo,
        should_retry: bool,
    ):
        if command.failed_nodes is None:
            command.failed_nodes = {}

        return (
            False  # todo: command.failed_nodes[chosen_node] = self.__read_exception_from_server(request, response, e)
        )

        if node_index is None:
            # We executed request over a node not in the topology. This means no failover...
            return False

        if self._node_selector is None:
            # todo: spawnHealthChecks(chosenNode, nodeIndex)
            return False

        # As the server is down, we discard the server version to ensure we update when it goes up.
        chosen_node.discard_server_version()

        self._node_selector.on_failed_request(node_index)

        if self.should_broadcast(command):
            command.result = self.__broadcast(command, session_info)
            return True

        # todo: self.spawn_health_checks(chosen_node, node_index)

        index_node_and_etag = self._node_selector.get_preferred_node_with_topology()
        if command.failover_topology_etag != self.topology_etag:
            command.failed_nodes.clear()
            command.failover_topology_etag = self.topology_etag

        if index_node_and_etag.current_node in command.failed_nodes:
            return False

        self.__on_failed_request_invoke(url, e, request, response)

        self.execute(
            index_node_and_etag.current_node, index_node_and_etag.current_index, command, should_retry, session_info
        )

        return True

    @staticmethod
    def __read_exception_from_server(request: requests.Request, response: requests.Response, e: Exception) -> Exception:
        if response and response.content:
            response_json = None
            try:
                response_json = response.content.decode("utf-8")

                # todo: change this bs
                def exception_schema_decoder(dictionary: dict) -> ExceptionDispatcher.ExceptionSchema:
                    return ExceptionDispatcher.ExceptionSchema(
                        dictionary.get("url"),
                        dictionary.get("class"),
                        dictionary.get("message"),
                        dictionary.get("error"),
                    )

                return ExceptionDispatcher.get(
                    json.loads(response_json, object_hook=exception_schema_decoder), response.status_code, e
                )
            except:
                exception_schema = ExceptionDispatcher.ExceptionSchema(
                    request.url,
                    "Unparsable Server Response",
                    "Get unrecognized response from the server",
                    response_json,
                )

                return ExceptionDispatcher.get(exception_schema, response.status_code, e)

        exception_schema = ExceptionDispatcher.ExceptionSchema(
            request.url,
            e.__class__.__qualname__,
            e.args[0],
            f"An exception occurred while contacting {request.url}.{os.linesep}{str(e)}",
        )
        return ExceptionDispatcher.get(exception_schema, HTTPStatus.SERVICE_UNAVAILABLE, e)

    class IndexAndResponse:
        def __init__(self, index: int, response: requests.Response):
            self.index = index
            self.response = response

    @client_configuration_etag.setter
    def client_configuration_etag(self, value):
        self._client_configuration_etag = value

    def _ensure_node_selector(self) -> None:
        if not self._disable_topology_updates:
            self.__wait_for_topology_update(self._first_topology_update_task)

        if self._node_selector is None:
            topology = Topology(self.topology_etag, self.topology_nodes)
            self._node_selector = NodeSelector(topology, self._thread_pool_executor)


class ClusterRequestExecutor(RequestExecutor):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.close()

    # todo: security, cryptography, https
    def __init__(
        self,
        conventions: DocumentConventions,
        certificate_path: Optional[str] = None,
        trust_store_path: Optional[str] = None,
        thread_pool_executor: Optional[ThreadPoolExecutor] = None,
        initial_urls: Optional[List[str]] = None,
    ):
        super(ClusterRequestExecutor, self).__init__(
            None, conventions, certificate_path, trust_store_path, thread_pool_executor, initial_urls
        )
        self.__cluster_topology_semaphore = Semaphore(1)

    @classmethod
    def create_for_single_node(
        cls,
        url: str,
        thread_pool_executor: ThreadPoolExecutor,
        conventions: Optional[DocumentConventions] = None,
        certificate_path: Optional[str] = None,
        trust_store_path: Optional[str] = None,
    ) -> ClusterRequestExecutor:
        initial_urls = [url]
        # todo: validate urls
        # todo: use default, static DocumentConventions
        executor = cls(
            (conventions if conventions else DocumentConventions()),
            certificate_path,
            trust_store_path,
            thread_pool_executor,
            initial_urls,
        )

        server_node = ServerNode(url)
        server_node.url = url

        topology = Topology(-1, [server_node])

        node_selector = NodeSelector(topology, thread_pool_executor)
        executor._node_selector = node_selector
        executor._topology_etag = -2
        executor._disable_client_configuration_updates = True
        executor._disable_topology_updates = True

        return executor

    @classmethod
    def create_without_database_name(
        cls,
        initial_urls: List[str],
        thread_pool_executor: ThreadPoolExecutor,
        conventions: DocumentConventions,
        certificate_path: Optional[str] = None,
        trust_store_path: Optional[str] = None,
    ) -> ClusterRequestExecutor:
        executor = cls(
            (conventions if conventions else DocumentConventions()),
            certificate_path,
            trust_store_path,
            thread_pool_executor,
            initial_urls,
        )
        executor._disable_client_configuration_updates = True
        executor._first_topology_update_task = executor._first_topology_update(initial_urls, None)
        return executor

    def update_topology_async(self, parameters: UpdateTopologyParameters) -> Future:
        if parameters is None:
            raise ValueError("Parameters cannot be None")

        if self._disposed:
            returned = Future()
            returned.set_result(False)
            return returned

        def __supply_async():
            lock_taken = self.__cluster_topology_semaphore.acquire(timeout=parameters.timeout_in_ms * 1000)
            if not lock_taken:
                return False
            try:
                if self._disposed:
                    return False

                command = GetClusterTopologyCommand(parameters.debug_tag)
                self.execute(parameters.node, None, command, False, None)

                results = command.result
                nodes = [
                    ServerNode(url=url, cluster_tag=cluster_tag)
                    for cluster_tag, url in results.topology.members.items()
                ]
                new_topology = Topology(results.etag, nodes)
                self._topology_etag = results.etag

                if self._node_selector is None:
                    self._node_selector = NodeSelector(new_topology, self._thread_pool_executor)

                    if self.conventions.read_balance_behavior == ReadBalanceBehavior.FASTEST_NODE:
                        self._node_selector.schedule_speed_test()

                elif self._node_selector.on_update_topology(new_topology, parameters.force_update):
                    self._dispose_all_failed_nodes_timers()

                    if self.conventions.read_balance_behavior == ReadBalanceBehavior.FASTEST_NODE:
                        self._node_selector.schedule_speed_test()

                self._on_topology_updated_invoke(new_topology)
            except BaseException as e:
                if not self._disposed:
                    raise e

            finally:
                self.__cluster_topology_semaphore.release()

            return True

        return self._thread_pool_executor.submit(__supply_async)

    def _throw_exceptions(self, details: str):
        raise RuntimeError(f"Failed to retrieve cluster topology from all known nodes {os.linesep}{details}")



import os
from typing import Optional, Type, TypeVar, Dict, List, TYPE_CHECKING

from ravendb.documents.operations.ongoing_tasks import ToggleOngoingTaskStateOperation, OngoingTaskType
from ravendb.documents.session.tokens.query_tokens.definitions import CounterIncludesToken, TimeSeriesIncludesToken
from ravendb.documents.session.tokens.query_tokens.query_token import QueryToken
from ravendb.documents.session.utils.includes_util import IncludesUtil
from ravendb.documents.commands.subscriptions import (
    CreateSubscriptionCommand,
    GetSubscriptionsCommand,
    DeleteSubscriptionCommand,
    GetSubscriptionStateCommand,
    DropSubscriptionConnectionCommand,
    UpdateSubscriptionCommand,
)
from ravendb.documents.session.loaders.include import SubscriptionIncludeBuilder
from ravendb.documents.subscriptions.options import (
    SubscriptionCreationOptions,
    SubscriptionWorkerOptions,
    SubscriptionUpdateOptions,
)
from ravendb.documents.subscriptions.revision import Revision
from ravendb.documents.subscriptions.state import SubscriptionState
from ravendb.documents.subscriptions.worker import SubscriptionWorker
from ravendb.extensions.string_extensions import escape_string

_T = TypeVar("_T")

if TYPE_CHECKING:
    from ravendb.documents.store.definition import DocumentStore


class DocumentSubscriptions:
    def __init__(self, store: "DocumentStore"):
        self._store = store
        self._subscriptions: Dict[SubscriptionWorker, bool] = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def create_for_options(self, options: SubscriptionCreationOptions, database: Optional[str] = None) -> str:
        if options is None:
            raise ValueError("Cannot create a subscription if options is None")

        if options.query is None:
            raise ValueError("Cannot create a subscriptions if the script is None")

        request_executor = self._store.get_request_executor(self._store.get_effective_database(database))

        command = CreateSubscriptionCommand(self._store.conventions, options)
        request_executor.execute_command(command)

        return command.result.name

    def create_for_options_autocomplete_query(
        self,
        object_type: Type[_T],
        options: SubscriptionCreationOptions = SubscriptionCreationOptions(),
        database: Optional[str] = None,
    ) -> str:
        options = self.ensure_criteria(options, object_type, False)
        return self.create_for_options(options, database)

    def create_for_class(
        self,
        object_type: Type[_T],
        options: Optional[SubscriptionCreationOptions] = None,
        database: Optional[str] = None,
    ) -> str:
        options = options or SubscriptionCreationOptions()
        return self.create_for_options(self.ensure_criteria(options, object_type, False), database)

    def ensure_criteria(self, criteria: SubscriptionCreationOptions, object_type: Type[_T], revisions: bool):
        if criteria is None:
            criteria = SubscriptionCreationOptions()

        collection_name = self._store.conventions.get_collection_name(object_type)

        if criteria.query:
            query_builder = [criteria.query]
        else:
            query_builder = ["from '"]
            escape_string(query_builder, collection_name)
            query_builder.append("'")

            if revisions:
                query_builder.append(" (Revisions = true)")

            query_builder.append(" as doc")

        if criteria.includes:
            builder = SubscriptionIncludeBuilder(self._store.conventions)
            criteria.includes(builder)

            number_of_includes_added = 0

            if builder.documents_to_include is not None and not len(builder.documents_to_include) == 0:
                query_builder.append(os.linesep)
                query_builder.append("include ")

                for inc in builder.documents_to_include:
                    include = "doc." + inc
                    if number_of_includes_added > 0:
                        query_builder.append(",")

                    escaped_include = None
                    req, escaped_include = IncludesUtil.requires_quotes(include)
                    if req:
                        query_builder.append("'")
                        query_builder.append(escaped_include)
                        query_builder.append("'")
                    else:
                        query_builder.append(f"'{include}'" if QueryToken.is_keyword(include) else include)

                    number_of_includes_added += 1

            if builder.is_all_counters:
                if number_of_includes_added == 0:
                    query_builder.append(os.linesep)
                    query_builder.append("include ")

                token = CounterIncludesToken.all("")
                token.write_to(query_builder)
                number_of_includes_added += 1

            elif builder.counters_to_include:
                if number_of_includes_added:
                    query_builder.append(os.linesep)
                    query_builder.append("include ")

                for counter_name in builder.counters_to_include:
                    if number_of_includes_added > 0:
                        query_builder.append(",")

                    token = CounterIncludesToken.create("", counter_name)
                    token.write_to(query_builder)

                    number_of_includes_added += 1

            if builder.time_series_to_include:
                for time_series_range in builder.time_series_to_include:
                    if number_of_includes_added == 0:
                        query_builder.append(os.linesep)
                        query_builder.append("include ")

                    if number_of_includes_added > 0:
                        query_builder.append(",")

                    token = TimeSeriesIncludesToken.create("", time_series_range)
                    token.write_to(query_builder)

        criteria.query = "".join(query_builder)
        return criteria

    def get_subscription_worker(
        self, options: SubscriptionWorkerOptions, object_type: Optional[Type[_T]] = None, database: Optional[str] = None
    ) -> SubscriptionWorker[_T]:
        self._store.assert_initialized()
        if options is None:
            raise RuntimeError("Cannot open a subscription if options are None")

        subscription = SubscriptionWorker(object_type, options, False, self._store, database)

        subscription._on_closed = lambda sender: self._subscriptions.pop(sender)
        self._subscriptions[subscription] = True

        return subscription

    def get_subscription_worker_by_name(
        self,
        subscription_name: Optional[str] = None,
        object_type: Optional[Type[_T]] = None,
        database: Optional[str] = None,
    ) -> SubscriptionWorker[_T]:
        return self.get_subscription_worker(SubscriptionWorkerOptions(subscription_name), object_type, database)

    def get_subscription_worker_for_revisions(
        self, options: SubscriptionWorkerOptions, object_type: Optional[Type[_T]] = None, database: Optional[str] = None
    ) -> SubscriptionWorker[Revision[_T]]:
        subscription = SubscriptionWorker(object_type, options, True, self._store, database)

        subscription._on_closed = lambda sender: self._subscriptions.pop(sender)
        self._subscriptions[subscription] = True

        return subscription

    def get_subscriptions(self, start: int, take: int, database: Optional[str] = None) -> List[SubscriptionState]:
        request_executor = self._store.get_request_executor(self._store.get_effective_database(database))

        command = GetSubscriptionsCommand(start, take)
        request_executor.execute_command(command)

        return command.result

    def delete(self, name: str, database: Optional[str] = None) -> None:
        request_executor = self._store.get_request_executor(self._store.get_effective_database(database))

        command = DeleteSubscriptionCommand(name)
        request_executor.execute_command(command)

    def get_subscription_state(self, subscription_name: str, database: Optional[str] = None) -> SubscriptionState:
        if not subscription_name or subscription_name.isspace():
            raise ValueError("Subscription name cannot be None")

        request_executor = self._store.get_request_executor(self._store.get_effective_database(database))

        command = GetSubscriptionStateCommand(subscription_name)
        request_executor.execute_command(command)
        return command.result

    def close(self) -> None:
        if not self._subscriptions:
            return

        for subscription in self._subscriptions:
            subscription.close()

    def drop_connection(self, name: str, database: Optional[str] = None) -> None:
        request_executor = self._store.get_request_executor(self._store.get_effective_database(database))

        command = DropSubscriptionConnectionCommand(name)
        request_executor.execute_command(command)

    def enable(self, name: str, database: Optional[str] = None) -> None:
        operation = ToggleOngoingTaskStateOperation(name, OngoingTaskType.SUBSCRIPTION, False)
        self._store.maintenance.for_database(self._store.get_effective_database(database)).send(operation)

    def disable(self, name: str, database: Optional[str] = None) -> None:
        operation = ToggleOngoingTaskStateOperation(name, OngoingTaskType.SUBSCRIPTION, True)
        self._store.maintenance.for_database(self._store.get_effective_database(database)).send(operation)

    def update(self, options: SubscriptionUpdateOptions, database: Optional[str] = None) -> str:
        if options is None:
            raise ValueError("Cannot update a subscription if options is None")

        if not options.name and options.key is None:
            raise ValueError("Cannot update a subscription if both options.name nad options.id are None")

        request_executor = self._store.get_request_executor(database)
        command = UpdateSubscriptionCommand(options)
        request_executor.execute_command(command, None)

        return command.result.name



import json
from typing import Dict

from ravendb.json.metadata_as_dictionary import MetadataAsDictionary
from ravendb.tools.utils import Utils


def _update_metadata_modifications(metadata_dictionary: MetadataAsDictionary, metadata: Dict) -> bool:
    dirty = False
    if metadata_dictionary is not None:
        if metadata_dictionary.is_dirty:
            dirty = True
        for key, value in metadata_dictionary.items():
            if value is None or isinstance(value, MetadataAsDictionary) and value.is_dirty is True:
                dirty = True
            metadata[key] = json.loads(json.dumps(value, default=Utils.json_default))
        if len(metadata) != len(metadata_dictionary):
            # looks like some props were removed
            to_remove = set()

            fields = metadata.keys()
            for field in fields:
                if field not in metadata_dictionary:
                    to_remove.add(field)

            for s in to_remove:
                del metadata[s]

    return dirty



from __future__ import annotations

import datetime
import enum
from typing import List, Optional, Dict, TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ravendb.serverwide.operations.common import DatabasePromotionStatus


class DocumentsCompressionConfiguration:
    def __init__(
        self,
        compress_revisions: Optional[bool] = None,
        compress_all_collections: Optional[bool] = None,
        collections: Optional[List[str]] = None,
    ):
        self.compress_revisions = compress_revisions
        self.compress_all_collections = compress_all_collections
        self.collections = collections

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> DocumentsCompressionConfiguration:
        return cls(json_dict["CompressRevisions"], json_dict["CompressAllCollections"], json_dict["Collections"])

    def to_json(self) -> Dict[str, Any]:
        return {
            "CompressRevisions": self.compress_revisions,
            "CompressAllCollections": self.compress_all_collections,
            "Collections": self.collections,
        }


class DeletionInProgressStatus(enum.Enum):
    NO = "NO"
    SOFT_DELETE = "SOFT_DELETE"
    HARD_DELETE = "HARD_DELETE"

    def __str__(self):
        return self.value


class ConflictSolver:
    def __init__(
        self,
        resolve_by_collection: Optional[Dict[str, ScriptResolver]] = None,
        resolve_to_latest: Optional[bool] = None,
    ):
        self.resolve_by_collection = resolve_by_collection
        self.resolve_to_latest = resolve_to_latest


class ScriptResolver:
    def __init__(self, script: Optional[str] = None, last_modified_time: Optional[datetime.datetime] = None):
        self.script = script
        self.last_modified_time = last_modified_time


class DatabaseTopology:
    def __init__(
        self,
        members: Optional[List[str]] = None,
        promotables: Optional[List[str]] = None,
        rehabs: Optional[List[str]] = None,
        predefined_mentors: Optional[Dict[str, str]] = None,
        demotion_reasons: Optional[Dict[str, str]] = None,
        promotables_status: Optional[Dict[str, "DatabasePromotionStatus"]] = None,
        replication_factor: Optional[int] = None,
        dynamic_nodes_distribution: Optional[bool] = None,
        stamp: Optional[LeaderStamp] = None,
        database_topology_id_base_64: Optional[str] = None,
        priority_order: Optional[List[str]] = None,
        nodes_modified_at: Optional[datetime.datetime] = None,
    ):
        self.members = members
        self.promotables = promotables
        self.rehabs = rehabs
        self.predefined_mentors = predefined_mentors
        self.demotion_reasons = demotion_reasons
        self.promotables_status = promotables_status
        self.replication_factor = replication_factor
        self.dynamic_nodes_distribution = dynamic_nodes_distribution
        self.stamp = stamp
        self.database_topology_id_base_64 = database_topology_id_base_64
        self.priority_order = priority_order
        self.nodes_modified_at = nodes_modified_at

    @property
    def all_nodes(self) -> List[str]:
        return [*self.members, *self.promotables, *self.rehabs]

    @classmethod
    def from_json(cls, json_dict: dict) -> DatabaseTopology:
        return cls(
            json_dict["Members"],
            json_dict["Promotables"],
            json_dict["Rehabs"],
            json_dict.get("PredefinedMentors", None),
            json_dict["DemotionReasons"],
            json_dict["PromotablesStatus"],
            json_dict["ReplicationFactor"],
            json_dict["DynamicNodesDistribution"],
            LeaderStamp.from_json(json_dict["Stamp"]),
            json_dict["DatabaseTopologyIdBase64"],
            json_dict["PriorityOrder"],
        )


class LeaderStamp:
    def __init__(self, index: Optional[int] = None, term: Optional[int] = None, leaders_ticks: Optional[int] = None):
        self.index = index
        self.term = term
        self.leaders_ticks = leaders_ticks

    @classmethod
    def from_json(cls, json_dict: Dict) -> LeaderStamp:
        return cls(json_dict["Index"], json_dict["Term"], json_dict["LeadersTicks"])


class CompactSettings:
    def __init__(self, database_name: str = None, documents: bool = None, indexes: List[str] = None):
        self.database_name = database_name
        self.documents = documents
        self.indexes = indexes

    def to_json(self) -> Dict[str, Any]:
        return {"DatabaseName": self.database_name, "Documents": self.documents, "Indexes": self.indexes}



from __future__ import annotations
import datetime
import enum
import re
from enum import Enum
from abc import ABC
from typing import Union, Optional, List, Dict, Set, Iterable
from ravendb.documents.indexes.spatial.configuration import SpatialOptions, AutoSpatialOptions
from ravendb.documents.indexes.vector.options import VectorOptions, AutoVectorOptions
from ravendb.tools.utils import Utils


# The sort options to use for a particular field
class SortOptions(Enum):
    # No sort options
    none = "None"
    # Sort using term values as Strings. Sort values are str and lower values are at the front.
    str = "String"
    # Sort using term values as encoded Doubles and Longs.
    # Sort values are float or longs and lower values are at the front.
    numeric = "Numeric"

    def __str__(self):
        return self.value


class IndexLockMode(Enum):
    UNLOCK = "Unlock"
    LOCKED_IGNORE = "LockedIgnore"
    LOCKED_ERROR = "LockedError"
    SIDE_BY_SIDE = "SideBySide"

    def __str__(self):
        return self.value


class IndexState(Enum):
    NORMAL = "Normal"
    DISABLED = "Disabled"
    IDLE = "Idle"
    ERROR = "Error"

    def __str__(self):
        return self.value


class IndexPriority(Enum):
    LOW = "Low"
    NORMAL = "Normal"
    HIGH = "High"

    def __str__(self):
        return self.value


class IndexDeploymentMode(Enum):
    PARALLEL = "Parallel"
    ROLLING = "Rolling"

    def __str__(self):
        return self.value


class SearchEngineType(Enum):
    LUCENE = "Lucene"
    CORAX = "Corax"

    def __str__(self):
        return self.value


class FieldStorage(Enum):
    YES = "Yes"
    NO = "No"


class FieldIndexing(Enum):
    NO = "No"
    SEARCH = "Search"
    EXACT = "Exact"
    HIGHLIGHTING = "Highlighting"
    DEFAULT = "Default"

    def __str__(self):
        return self.value


class AutoFieldIndexing(Enum):
    NO = "No"
    SEARCH = "Search"
    EXACT = "Exact"
    HIGHLIGHTING = "Highlighting"
    DEFAULT = "Default"

    def __str__(self):
        return self.value


class FieldTermVector(Enum):
    YES = "Yes"
    NO = "No"
    WITH_POSITIONS = "WithPositions"
    WITH_OFFSETS = "WithOffsets"
    WITH_POSITIONS_AND_OFFSETS = "WithPositionsAndOffsets"

    def __str__(self):
        return self.value


class IndexSourceType(Enum):
    NONE = "None"
    DOCUMENTS = "Documents"
    TIME_SERIES = "TimeSeries"
    COUNTERS = "Counters"

    def __str__(self):
        return self.value


class IndexType(Enum):
    NONE = "None"
    AUTO_MAP = "AutoMap"
    AUTO_MAP_REDUCE = "AutoMapReduce"
    MAP = "Map"
    MAP_REDUCE = "MapReduce"
    FAULTY = "Faulty"
    JAVA_SCRIPT_MAP = "JavaScriptMap"
    JAVA_SCRIPT_MAP_REDUCE = "JavaScriptMapReduce"

    def __str__(self):
        return self.value


class AggregationOperation(Enum):
    NONE = "None"
    COUNT = "Count"
    SUM = "Sum"

    def __str__(self):
        return self.value


class GroupByArrayBehavior(Enum):
    NOT_APPLICABLE = "NotApplicable"
    BY_CONTENT = "ByContent"
    BY_INDIVIDUAL_VALUES = "ByIndividualValues"


class IndexFieldOptions:
    def __init__(
        self,
        storage: Optional[FieldStorage] = None,
        indexing: Optional[FieldIndexing] = None,
        term_vector: Optional[FieldTermVector] = None,
        spatial: Optional[SpatialOptions] = None,
        vector: Optional[VectorOptions] = None,
        analyzer: Optional[str] = None,
        suggestions: Optional[bool] = None,
    ):
        self.storage = storage
        self.indexing = indexing
        self.term_vector = term_vector
        self.spatial = spatial
        self.vector = vector
        self.analyzer = analyzer
        self.suggestions = suggestions

    def to_json(self):
        return {
            "Storage": self.storage,
            "Indexing": self.indexing,
            "TermVector": self.term_vector,
            "Spatial": self.spatial.to_json() if self.spatial else None,
            "Vector": self.vector.to_json() if self.vector else None,
            "Analyzer": self.analyzer,
            "Suggestions": self.suggestions,
        }


class IndexDefinitionBase:
    def __init__(self, name: str = None, priority: IndexPriority = None, state: IndexState = None):
        self.name = name
        self.priority = priority
        self.state = state


class IndexDefinition(IndexDefinitionBase):
    def __init__(
        self,
        name: Optional[str] = None,
        priority: Optional[str] = None,
        state: Optional[str] = None,
        lock_mode: Optional[IndexLockMode] = None,
        additional_sources: Optional[Dict[str, str]] = None,
        additional_assemblies: Optional[Set[AdditionalAssembly]] = None,
        maps: Optional[Set[str]] = None,
        fields: Optional[Dict[str, IndexFieldOptions]] = None,
        reduce: Optional[str] = None,
        configuration: Optional[Dict[str, str]] = None,
        index_source_type: Optional[IndexSourceType] = None,
        index_type: Optional[IndexType] = None,
        output_reduce_to_collection: Optional[str] = None,
        reduce_output_index: Optional[int] = None,
        pattern_for_output_reduce_to_collection_references: Optional[str] = None,
        pattern_references_collection_name: Optional[str] = None,
        deployment_mode: Optional[IndexDeploymentMode] = None,
        search_engine_type: Optional[SearchEngineType] = None,
    ):
        super(IndexDefinition, self).__init__(name, priority, state)
        self.lock_mode = lock_mode
        self.additional_sources = additional_sources
        self.additional_assemblies = additional_assemblies
        self.__maps = maps or set()
        self.fields = fields or {}
        self.reduce = reduce
        self.configuration = configuration or {}
        self.__index_source_type = index_source_type
        self.__index_type = index_type
        self.output_reduce_to_collection = output_reduce_to_collection
        self.reduce_output_index = reduce_output_index
        self.pattern_for_output_reduce_to_collection_references = pattern_for_output_reduce_to_collection_references
        self.pattern_references_collection_name = pattern_references_collection_name
        self.deployment_mode = deployment_mode
        self.search_engine_type = search_engine_type

    @classmethod
    def from_json(cls, json_dict: dict) -> IndexDefinition:
        result = cls()
        result.name = json_dict["Name"]
        result.priority = IndexPriority(json_dict["Priority"])
        result.state = IndexState(json_dict["State"])
        result.lock_mode = IndexLockMode(json_dict["LockMode"])
        result.additional_sources = json_dict["AdditionalSources"]
        result.additional_assemblies = set(map(AdditionalAssembly, json_dict["AdditionalAssemblies"]))
        result.maps = json_dict["Maps"]
        result.fields = map(
            lambda key, value: {key: IndexFieldOptions(value)}, json_dict["Fields"]
        )  # todo: do also nested types taken in IndexFieldOptions init
        result.reduce = json_dict["Reduce"]
        result.configuration = json_dict["Configuration"]
        source_type = json_dict.get("IndexSourceType", None)
        if source_type is not None:
            result.__index_source_type = IndexSourceType(source_type)
        index_type = json_dict.get("IndexType", None)
        if index_type is not None:
            result.__index_type = IndexType(index_type)
        if json_dict["Configuration"] and "Indexing.Static.SearchEngineType" in json_dict["Configuration"]:
            result.search_engine_type = SearchEngineType(json_dict["Configuration"]["Indexing.Static.SearchEngineType"])
        result.output_reduce_to_collection = json_dict["OutputReduceToCollection"]
        result.reduce_output_index = json_dict["ReduceOutputIndex"]
        result.pattern_for_output_reduce_to_collection_references = json_dict[
            "PatternForOutputReduceToCollectionReferences"
        ]
        result.pattern_references_collection_name = json_dict["PatternReferencesCollectionName"]
        deploy = json_dict.get("DeploymentMode", None)
        if deploy is not None:
            result.deployment_mode = IndexDeploymentMode(deploy)
        return result

    def to_json(self) -> dict:
        return {
            "Name": self.name,
            "Priority": self.priority,
            "State": self.state,
            "LockMode": self.lock_mode,
            "AdditionalSources": self.additional_sources,
            "AdditionalAssemblies": (
                [x.to_json() for x in self.additional_assemblies] if self.additional_assemblies else None
            ),
            "Maps": self.maps,
            "Fields": {key: value.to_json() for key, value in self.fields.items()},
            "Reduce": self.reduce,
            "Configuration": self.configuration,
            "IndexSourceType": self.__index_source_type,
            "IndexType": self.__index_type,
            "OutputReduceToCollection": self.output_reduce_to_collection,
            "ReduceOutputIndex": self.reduce_output_index,
            "PatternForOutputReduceToCollectionReferences": self.pattern_for_output_reduce_to_collection_references,
            "PatternReferencesCollectionName": self.pattern_references_collection_name,
            "DeploymentMode": self.deployment_mode,
        }

    @property
    def maps(self):
        if self.__maps is None:
            self.__maps = set()
        return self.__maps

    @maps.setter
    def maps(self, value: Union[str, Iterable[str]]):
        if isinstance(value, str):
            self.__maps = {value}
        else:
            self.__maps = set(value)

    @property
    def source_type(self) -> IndexSourceType:
        if self.__index_source_type is None or self.__index_source_type == IndexSourceType.NONE:
            self.__index_source_type = self.detect_static_index_source_type()
        return self.__index_source_type

    @source_type.setter
    def source_type(self, value: IndexSourceType):
        self.__index_source_type = value

    @property
    def type(self) -> IndexType:
        if self.__index_type is None or self.__index_type == IndexType.NONE:
            self.__index_type = self.detect_static_index_type()
        return self.__index_type

    @type.setter
    def type(self, value: IndexType):
        self.__index_type = value

    def detect_static_index_source_type(self) -> IndexSourceType:
        if not self.maps:
            raise ValueError("Index definition contains no maps")

        source_type = IndexSourceType.NONE
        for map in self.maps:
            map_source_type = IndexDefinitionHelper.detect_static_index_source_type(map)
            if source_type == IndexSourceType.NONE:
                source_type = map_source_type
                continue

            if source_type != map_source_type:
                raise RuntimeError("Index definition cannot contain maps with different source types.")

        return source_type

    def detect_static_index_type(self) -> IndexType:
        first_map = self.maps.__iter__().__next__()

        if first_map is None:
            raise ValueError("Index definitions contains no Maps")

        return None  # todo: IndexDefinitionHelper.detect_Static_index_type(first_map, self.reduce)


class AutoIndexDefinition(IndexDefinitionBase):
    def __init__(
        self,
        index_type: Optional[IndexType] = None,
        name: Optional[str] = None,
        priority: Optional[IndexPriority] = None,
        state: Optional[IndexState] = None,
        collection: Optional[str] = None,
        map_fields: Optional[Dict[str, AutoIndexFieldOptions]] = None,
        group_by_fields: Optional[Dict[str, AutoIndexFieldOptions]] = None,
    ):
        super(AutoIndexDefinition, self).__init__(name, priority, state)
        self.index_type = index_type
        self.collection = collection
        self.map_fields = map_fields
        self.group_by_fields = group_by_fields

    @classmethod
    def from_json(cls, json_dict: Dict) -> AutoIndexDefinition:
        return cls(
            IndexType(json_dict.get("Type")),
            json_dict.get("Name"),
            IndexPriority(json_dict.get("Priority")),
            IndexState(json_dict.get("State")) if json_dict.get("State", None) else None,
            json_dict.get("Collection"),
            {name: AutoIndexFieldOptions.from_json(value) for name, value in json_dict.get("MapFields").items()},
            {name: AutoIndexFieldOptions.from_json(value) for name, value in json_dict.get("GroupByFields").items()},
        )

    def to_json(self) -> Dict:
        return {
            "Type": self.index_type.value,
            "Name": self.name,
            "Priority": self.priority.value,
            "State": self.state.value if self.state is not None else None,
            "Collection": self.collection,
            "MapFields": {key: map_field.to_json() for key, map_field in self.map_fields.items()},
            "GroupByFields": {key: group_by_field.to_json() for key, group_by_field in self.group_by_fields.items()},
        }


class AutoIndexFieldOptions:
    def __init__(
        self,
        storage: Optional[FieldStorage] = None,
        indexing: Optional[AutoFieldIndexing] = None,
        aggregation: Optional[AggregationOperation] = None,
        spatial: Optional[AutoSpatialOptions] = None,
        vector: Optional[AutoVectorOptions] = None,
        group_by_array_behavior: Optional[GroupByArrayBehavior] = None,
        suggestions: Optional[bool] = None,
        is_name_quoted: Optional[bool] = None,
    ):
        self.storage = storage
        self.indexing = indexing
        self.aggregation = aggregation
        self.spatial = spatial
        self.vector = vector
        self.group_by_array_behavior = group_by_array_behavior
        self.suggestions = suggestions
        self.is_name_quoted = is_name_quoted

    @classmethod
    def from_json(cls, json_dict: Dict) -> AutoIndexFieldOptions:
        return cls(
            FieldStorage(json_dict.get("Storage")),
            AutoFieldIndexing(json_dict.get("Indexing")),
            AggregationOperation(json_dict.get("Aggregation")) if json_dict.get("Aggregation", None) else None,
            AutoSpatialOptions.from_json(json_dict.get("Spatial")) if json_dict.get("Spatial", None) else None,
            AutoVectorOptions.from_json(json_dict.get("Vector")) if json_dict.get("Vector", None) else None,
            GroupByArrayBehavior(json_dict.get("GroupByArrayBehavior")),
            json_dict.get("Suggestions"),
            json_dict.get("IsNameQuoted"),
        )

    def to_json(self) -> Dict:
        return {
            "Storage": self.storage.value,
            "Indexing": self.indexing.value,
            "Aggregation": self.aggregation.value if self.aggregation is not None else None,
            "Spatial": self.spatial.type if self.spatial is not None else None,
            "Vector": (
                self.vector.to_json() if self.vector is not None else None
            ),  # todo; check if vector.to_json() is valid here
            "GroupByArrayBehavior": self.group_by_array_behavior.value,
            "Suggestions": self.suggestions,
            "IsNameQuoted": self.is_name_quoted,
        }


class AdditionalAssembly:
    def __init__(
        self,
        assembly_name: Optional[str] = None,
        assembly_path: Optional[str] = None,
        package_name: Optional[str] = None,
        package_version: Optional[str] = None,
        package_source_url: Optional[str] = None,
        usings: Optional[Set[str]] = None,
    ):
        self.assembly_name = assembly_name
        self.assembly_path = assembly_path
        self.package_name = package_name
        self.package_version = package_version
        self.package_source_url = package_source_url
        self.usings = usings

    def to_json(self) -> dict:
        return {
            "AssemblyName": self.assembly_name,
            "AssemblyPath": self.assembly_path,
            "PackageName": self.package_name,
            "PackageVersion": self.package_version,
            "PackageSourceUrl": self.package_source_url,
            "Usings": self.usings,
        }

    @classmethod
    def only_usings(cls, usings: Set[str]) -> AdditionalAssembly:
        if not usings:
            raise ValueError("using cannot be None or empty")
        return cls(usings=usings)

    @classmethod
    def from_runtime(cls, assembly_name: str, usings: Optional[Set[str]] = None) -> AdditionalAssembly:
        if not assembly_name or assembly_name.isspace():
            raise ValueError("Assembly name cannot be None or whitespace")
        return cls(assembly_name, usings=usings)

    @classmethod
    def from_path(cls, assembly_path: str, usings: Optional[Set[str]] = None) -> AdditionalAssembly:
        if not assembly_path or assembly_path.isspace():
            raise ValueError("Assembly path cannot be None or whitespace")
        return cls(assembly_path=assembly_path, usings=usings)

    @classmethod
    def from_NuGet(
        cls,
        package_name: str,
        package_version: str,
        package_source_url: Optional[str] = None,
        usings: Optional[Set[str]] = None,
    ) -> AdditionalAssembly:
        if not package_name or package_name.isspace():
            raise ValueError("Package name cannot be None or whitespace")
        if not package_version or package_version.isspace():
            raise ValueError("Package version cannot be None or whitespace")

        return cls(
            package_name=package_name,
            package_version=package_version,
            package_source_url=package_source_url,
            usings=usings,
        )


class AbstractCommonApiForIndexes(ABC):
    def __init__(self):
        self.__additional_sources: Optional[Dict[str, str]] = None
        self.__additional_assemblies: Optional[Set[AdditionalAssembly]] = None
        self.__configuration: Optional[Dict[str, str]] = None

    @property
    def is_map_reduce(self) -> bool:
        return False

    @property
    def index_name(self) -> str:
        return type(self).__name__.replace("_", "/")

    @property
    def additional_sources(self) -> Dict[str, str]:
        return self.__additional_sources

    @additional_sources.setter
    def additional_sources(self, value: Dict[str, str]):
        self.__additional_sources = value

    @property
    def additional_assemblies(self):
        return self.__additional_assemblies

    @additional_assemblies.setter
    def additional_assemblies(self, value: Set[AdditionalAssembly]):
        self.__additional_assemblies = value

    @property
    def configuration(self) -> Dict[str, str]:
        return self.__configuration

    @configuration.setter
    def configuration(self, value: Dict[str, str]):
        self.__configuration = value


class RollingIndexState(Enum):
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    DONE = "DONE"

    def __str__(self):
        return self.value


class RollingIndex:
    def __init__(
        self, active_deployments: Optional[Dict[str, RollingIndexDeployment]], raft_command_index: Optional[int] = None
    ):
        self.active_deployments = active_deployments
        self.raft_command_index = raft_command_index


class RollingIndexDeployment:
    def __init__(
        self,
        state: Optional[RollingIndexState] = None,
        created_at: Optional[datetime.datetime] = None,
        started_at: Optional[datetime.datetime] = None,
        finished_at: Optional[datetime.datetime] = None,
    ):
        self.state = state
        self.created_at = created_at
        self.started_at = started_at
        self.finished_at = finished_at


class IndexingError:
    def __init__(
        self,
        error: Optional[str] = None,
        timestamp: Optional[datetime.datetime] = None,
        document: Optional[str] = None,
        action: Optional[str] = None,
    ):
        self.error = error
        self.timestamp = timestamp
        self.document = document
        self.action = action

    def __str__(self):
        return f"Error: {self.error}, Document: {self.document}, Action: {self.action}"

    @classmethod
    def from_json(cls, json_dict: dict) -> IndexingError:
        return cls(
            json_dict["Error"],
            Utils.string_to_datetime(json_dict["Timestamp"]),
            json_dict["Document"],
            json_dict["Action"],
        )


class IndexErrors:
    def __init__(self, name: Optional[str] = None, errors: Optional[List[IndexingError]] = None):
        self.name = name
        self.errors = errors

    @classmethod
    def from_json(cls, json_dict: dict) -> IndexErrors:
        return cls(json_dict["Name"], list(map(lambda x: IndexingError.from_json(x), json_dict["Errors"])))


class IndexDefinitionHelper:
    @staticmethod
    def detect_static_index_type(map_str: str, reduce: str) -> IndexType:
        if len(map_str) == 0:
            raise ValueError("Index definitions contains no maps")

        map_str = IndexDefinitionHelper._strip_comments(map_str)
        map_str = IndexDefinitionHelper._unify_white_space(map_str)

        map_lower = map_str.lower()
        if (
            map_lower.startswith("from")
            or map_lower.startswith("docs")
            or (map_lower.startswith("timeseries") and not map_lower.startswith("timeseries.map"))
            or (map_lower.startswith("counters") and not map_lower.startswith("counters.map"))
        ):
            # C# indexes must start with "from" query syntax or
            # "docs" for method syntax
            if reduce is None or reduce.isspace():
                return IndexType.MAP
            return IndexType.MAP_REDUCE

        if reduce.isspace():
            return IndexType.JAVA_SCRIPT_MAP

        return IndexType.JAVA_SCRIPT_MAP_REDUCE

    @staticmethod
    def detect_static_index_source_type(map_str: str) -> IndexSourceType:
        if not map_str or map_str.isspace():
            raise ValueError("Value cannot be None or whitespace")

        map_str = IndexDefinitionHelper._strip_comments(map_str)
        map_str = IndexDefinitionHelper._unify_white_space(map_str)

        # detect first supported syntax: timeseries.Companies.HeartRate.Where
        map_lower = map_str.lower()
        if map_lower.startswith("timeseries"):
            return IndexSourceType.TIME_SERIES

        if map_lower.startswith("counters"):
            return IndexSourceType.COUNTERS

        if map_lower.startswith("from"):
            # detect 'from ts in timeseries' or 'from ts in timeseries.Users.HeartRate'

            tokens = [token for token in map_lower.split(" ", 4) if token]

            if len(tokens) >= 4 and tokens[2].lower() == "in":
                if tokens[3].startswith("timeseries"):
                    return IndexSourceType.TIME_SERIES
                if tokens[3].startswith("counters"):
                    return IndexSourceType.COUNTERS

        # fallback to documents based index
        return IndexSourceType.DOCUMENTS

    @staticmethod
    def _strip_comments(map_str: str) -> str:
        return re.sub("(?:/\\*(?:[^*]|(?:\\*+[^*/]))*\\*+/)|(?://.*)", "", map_str).strip()

    @staticmethod
    def _unify_white_space(map_str: str) -> str:
        return re.sub("\\s+", " ", map_str)


class IndexRunningStatus(enum.Enum):
    RUNNING = "Running"
    PAUSED = "Paused"
    DISABLED = "Disabled"
    PENDING = "Pending"



from typing import List

from ravendb.documents.session.tokens.query_tokens.query_token import QueryToken
import ravendb.documents.session.tokens.query_tokens.definitions as tokens


class DocumentQueryHelper:
    @staticmethod
    def add_space_if_needed(previous_token: QueryToken, current_token: QueryToken, writer: List[str]) -> None:
        if previous_token is None:
            return

        if isinstance(previous_token, tokens.OpenSubclauseToken) or isinstance(
            current_token, (tokens.CloseSubclauseToken, tokens.IntersectMarkerToken)
        ):
            return

        writer.append(" ")



from abc import abstractmethod, ABC
from typing import TYPE_CHECKING, Generic, TypeVar, Optional

if TYPE_CHECKING:
    from ravendb.documents.commands.multi_get import GetRequest, GetResponse
    from ravendb.documents.queries.query import QueryResult

_T = TypeVar("_T")


class LazyOperation(Generic[_T], ABC):
    @abstractmethod
    def __init__(self):
        self.query_result: Optional["QueryResult"] = None
        self.result: _T = None

    @property
    @abstractmethod
    def requires_retry(self) -> bool:
        pass

    @abstractmethod
    def handle_response(self, response: "GetResponse") -> None:
        pass

    @abstractmethod
    def create_request(self) -> Optional["GetRequest"]:
        pass



from __future__ import annotations
import datetime
from typing import Optional, Dict

from ravendb.tools.utils import Utils


class SubscriptionState:
    def __init__(
        self,
        query: Optional[str] = None,
        change_vector_for_next_batch_starting_point: Optional[str] = None,
        subscription_id: Optional[int] = None,
        subscription_name: Optional[str] = None,
        mentor_node: Optional[str] = None,
        node_tag: Optional[str] = None,
        last_batch_ack_time: Optional[datetime.datetime] = None,
        last_client_connection_time: Optional[datetime.datetime] = None,
        disabled: Optional[bool] = None,
    ):
        self.query = query
        self.change_vector_for_next_batch_starting_point = change_vector_for_next_batch_starting_point
        self.subscription_id = subscription_id
        self.subscription_name = subscription_name
        self.mentor_node = mentor_node
        self.node_tag = node_tag
        self.last_batch_ack_time = last_batch_ack_time
        self.last_client_connection_time = last_client_connection_time
        self.disabled = disabled

    @classmethod
    def from_json(cls, json_dict: Dict) -> SubscriptionState:
        return cls(
            json_dict["Query"],
            json_dict["ChangeVectorForNextBatchStartingPoint"],
            json_dict["SubscriptionId"],
            json_dict["SubscriptionName"],
            json_dict.get("MentorNode", None),
            json_dict.get("NodeTag", None),
            Utils.string_to_datetime(json_dict["LastBatchAckTime"]),
            Utils.string_to_datetime(json_dict["LastClientConnectionTime"]),
            json_dict["Disabled"],
        )



from __future__ import annotations

import datetime
from typing import Optional, Dict, Any, List, Type, TypeVar, Generic

from ravendb.documents.session.time_series import (
    TimeSeriesEntry,
    TypedTimeSeriesEntry,
    TimeSeriesValuesHelper,
    ITimeSeriesValuesBindable,
)
from ravendb.tools.utils import Utils

_T_TS_Bindable = TypeVar("_T_TS_Bindable", bound=ITimeSeriesValuesBindable)


class TimeSeriesQueryResult:
    def __init__(self, count: Optional[int] = None):
        self.count = count

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> TimeSeriesQueryResult:
        return cls(json_dict["Count"])


class TimeSeriesRawResult(TimeSeriesQueryResult):
    def __init__(self, count: Optional[int] = None, results: Optional[List[TimeSeriesEntry]] = None):
        super().__init__(count)
        self.results = results

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> TimeSeriesQueryResult:
        if "Count" not in json_dict:
            json_dict = json_dict["__timeSeriesQueryFunction"]
        return cls(
            json_dict["Count"],
            [TimeSeriesEntry.from_json(time_series_entry_json) for time_series_entry_json in json_dict["Results"]],
        )

    def as_typed_result(self, object_type: Type[_T_TS_Bindable]) -> TypedTimeSeriesRawResult[_T_TS_Bindable]:
        result = TypedTimeSeriesRawResult()
        result.count = self.count
        result.results = [time_series_entry.as_typed_entry(object_type) for time_series_entry in self.results]
        return result


class TypedTimeSeriesRawResult(TimeSeriesQueryResult, Generic[_T_TS_Bindable]):
    def __init__(
        self, count: Optional[int] = None, results: Optional[List[TypedTimeSeriesEntry[_T_TS_Bindable]]] = None
    ):
        super().__init__(count)
        self.results = results

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> TimeSeriesQueryResult:
        return cls(
            json_dict["Count"],
            [TypedTimeSeriesEntry.from_json(typed_ts_entry_json) for typed_ts_entry_json in json_dict["Results"]],
        )


class TimeSeriesQueryBuilder:
    def __init__(self, query: str = None):
        self._query = query

    @property
    def query_text(self):
        return self._query

    def raw(self, query_text: str) -> TimeSeriesQueryResult:
        self._query = query_text
        return None


class TimeSeriesRangeAggregation:
    def __init__(
        self,
        count: List[int] = None,
        max: List[float] = None,
        min: List[float] = None,
        last: List[float] = None,
        first: List[float] = None,
        average: List[float] = None,
        sum: List[float] = None,
        to_date: datetime.datetime = None,
        from_date: datetime.datetime = None,
    ):
        self.count = count
        self.max = max
        self.min = min
        self.last = last
        self.first = first
        self.average = average
        self.sum = sum
        self.to_date = to_date
        self.from_date = from_date

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> TimeSeriesRangeAggregation:
        return cls(
            json_dict["Count"] if "Count" in json_dict else None,
            json_dict["Max"] if "Max" in json_dict else None,
            json_dict["Min"] if "Min" in json_dict else None,
            json_dict["Last"] if "Last" in json_dict else None,
            json_dict["First"] if "First" in json_dict else None,
            json_dict["Average"] if "Average" in json_dict else None,
            json_dict["Sum"] if "Sum" in json_dict else None,
            Utils.string_to_datetime(json_dict["To"]),
            Utils.string_to_datetime(json_dict["From"]),
        )

    def as_typed_entry(
        self, ts_bindable_object_type: Type[_T_TS_Bindable]
    ) -> TypedTimeSeriesRangeAggregation[_T_TS_Bindable]:
        typed_entry = TypedTimeSeriesRangeAggregation()

        typed_entry.from_date = self.from_date
        typed_entry.to_date = self.to_date
        typed_entry.min = (
            TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, self.min, False) if self.min else None
        )
        typed_entry.max = (
            TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, self.max, False) if self.max else None
        )
        typed_entry.first = (
            TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, self.first, False) if self.first else None
        )
        typed_entry.last = (
            TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, self.last, False) if self.last else None
        )
        typed_entry.sum = (
            TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, self.sum, False) if self.sum else None
        )
        counts = [float(count) for count in self.count]
        typed_entry.count = (
            TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, counts, False) if self.count else None
        )
        typed_entry.average = (
            TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, self.average, False) if self.average else None
        )

        return typed_entry


class TypedTimeSeriesRangeAggregation(Generic[_T_TS_Bindable]):
    def __init__(
        self,
        count: _T_TS_Bindable = None,
        max: _T_TS_Bindable = None,
        min: _T_TS_Bindable = None,
        last: _T_TS_Bindable = None,
        first: _T_TS_Bindable = None,
        average: _T_TS_Bindable = None,
        sum: _T_TS_Bindable = None,
        to_date: datetime.datetime = None,
        from_date: datetime.datetime = None,
    ):
        self.count = count
        self.max = max
        self.min = min
        self.last = last
        self.first = first
        self.average = average
        self.sum = sum
        self.to_date = to_date
        self.from_date = from_date


class TypedTimeSeriesAggregationResult(TimeSeriesQueryResult, Generic[_T_TS_Bindable]):
    def __init__(
        self, count: Optional[int] = None, results: List[TypedTimeSeriesRangeAggregation[_T_TS_Bindable]] = None
    ):
        super(TypedTimeSeriesAggregationResult, self).__init__(count)
        self.results = results


class TimeSeriesAggregationResult(TimeSeriesQueryResult):
    def __init__(self, count: Optional[int] = None, results: Optional[List[TimeSeriesRangeAggregation]] = None):
        super().__init__(count)
        self.results = results

    def as_typed_result(
        self, ts_bindable_object_type: Type[_T_TS_Bindable]
    ) -> TypedTimeSeriesAggregationResult[_T_TS_Bindable]:
        result = TypedTimeSeriesAggregationResult()
        result.count = self.count
        result.results = [x.as_typed_entry(ts_bindable_object_type) for x in self.results]
        return result

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> TimeSeriesAggregationResult:
        if "Count" not in json_dict:
            json_dict = json_dict["__timeSeriesQueryFunction"]
        return cls(
            json_dict["Count"], [TimeSeriesRangeAggregation.from_json(result) for result in json_dict["Results"]]
        )



from __future__ import annotations
import json
from typing import Optional, Generic, TypeVar, Dict, TYPE_CHECKING, Type, Collection

import requests

from ravendb.primitives import constants
from ravendb.documents.conventions import DocumentConventions
from ravendb.documents.operations.compare_exchange.compare_exchange import (
    CompareExchangeValue,
    CompareExchangeSessionValue,
)
from ravendb.documents.operations.compare_exchange.compare_exchange_value_result_parser import (
    CompareExchangeValueResultParser,
)
from ravendb.documents.operations.definitions import IOperation
from ravendb.http.http_cache import HttpCache
from ravendb.http.raven_command import RavenCommand
from ravendb.http.server_node import ServerNode
from ravendb.http.topology import RaftCommand
from ravendb.json.metadata_as_dictionary import MetadataAsDictionary
from ravendb.documents.session.entity_to_json import EntityToJsonStatic
from ravendb.tools.utils import Utils
from ravendb.util.util import RaftIdGenerator

_T = TypeVar("_T")

if TYPE_CHECKING:
    from ravendb.documents.store import DocumentStore


class CompareExchangeResult(Generic[_T]):
    def __init__(self, value: _T, index: int, successful: bool):
        self.value = value
        self.index = index
        self.successful = successful

    @classmethod
    def parse_from_string(
        cls, object_type: type, response_string: str, conventions: DocumentConventions
    ) -> CompareExchangeResult:
        response: dict = json.loads(response_string)

        index = response.get("Index", None)
        if index is None:
            raise RuntimeError("Response is invalid. Index is missing.")

        successful = response.get("Successful")
        raw = response.get("Value")

        val = None

        if raw:
            val = raw.get(constants.CompareExchange.OBJECT_FIELD_NAME)

        if val is None:
            return cls(Utils.json_default(object_type), index, successful)

        result = (
            Utils.convert_json_dict_to_object(val, object_type) if not isinstance(val, (str, int, float, bool)) else val
        )

        return cls(result, index, successful)


class PutCompareExchangeValueOperation(IOperation[CompareExchangeResult], Generic[_T]):
    def __init__(self, key: str, value: _T, index: int, metadata: MetadataAsDictionary = None):
        self._key = key
        self._value = value
        self._index = index
        self._metadata = metadata

    def get_command(self, store: DocumentStore, conventions: DocumentConventions, cache: HttpCache) -> RavenCommand[_T]:
        return self.__PutCompareExchangeValueCommand(self._key, self._value, self._index, self._metadata, conventions)

    class __PutCompareExchangeValueCommand(RavenCommand[_T], RaftCommand, Generic[_T]):
        def __init__(
            self, key: str, value: _T, index: int, metadata: MetadataAsDictionary, conventions: DocumentConventions
        ):
            if not key:
                raise ValueError("The key argument must have value")

            if index < 0:
                raise ValueError("Index must be a non-negative number")

            super().__init__(CompareExchangeResult[_T])

            self._key = key
            self._value = value
            self._index = index
            self._metadata = metadata
            self._conventions = conventions or DocumentConventions()

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{node.database}/cmpxchg?key={Utils.quote_key(self._key)}&index={self._index}"
            object_and_value = {constants.CompareExchange.OBJECT_FIELD_NAME: self._value}
            json_dict = EntityToJsonStatic.convert_entity_to_json(object_and_value, self._conventions, None, False)
            if self._metadata:
                metadata = CompareExchangeSessionValue.prepare_metadata_for_put(
                    self._key, self._metadata, self._conventions
                )
                json_dict[constants.Documents.Metadata.KEY] = metadata

            return requests.Request("PUT", url, data=json_dict)

        def set_response(self, response: str, from_cache: bool) -> None:
            self.result = CompareExchangeResult.parse_from_string(type(self._value), response, self._conventions)

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class GetCompareExchangeValueOperation(IOperation[CompareExchangeValue[_T]], Generic[_T]):
    def __init__(self, key: str, object_type: Type[_T], materialize_metadata: bool = True):
        self._key = key
        self._object_type = object_type
        self._materialize_metadata = materialize_metadata

    def get_command(self, store: DocumentStore, conventions: DocumentConventions, cache: HttpCache) -> RavenCommand[_T]:
        return self.GetCompareExchangeValueCommand(
            self._key, self._object_type, self._materialize_metadata, conventions
        )

    class GetCompareExchangeValueCommand(RavenCommand[CompareExchangeValue[_T]]):
        def __init__(
            self, key: str, object_type: Type[_T], materialize_metadata: bool, conventions: DocumentConventions
        ):
            if not key:
                raise ValueError("The key argument must have value")
            super().__init__(CompareExchangeValue[_T])
            self._key = key
            self._object_type = object_type
            self._materialize_metadata = materialize_metadata
            self._conventions = conventions

        def is_read_request(self) -> bool:
            return True

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{node.database}/cmpxchg?key={Utils.quote_key(self._key)}"
            return requests.Request("GET", url)

        def set_response(self, response: str, from_cache: bool) -> None:
            self.result = CompareExchangeValueResultParser.get_value(
                self._object_type, response, self._materialize_metadata, self._conventions
            )


class DeleteCompareExchangeValueOperation(IOperation[CompareExchangeResult[_T]], Generic[_T]):
    def __init__(self, object_type: type, key: str, index: int):
        self._key = key
        self._object_type = object_type
        self._index = index

    def get_command(self, store: DocumentStore, conventions: DocumentConventions, cache: HttpCache) -> RavenCommand[_T]:
        return self.RemoveCompareExchangeCommand[_T](self._object_type, self._key, self._index, conventions)

    class RemoveCompareExchangeCommand(RavenCommand[CompareExchangeResult[_T]], RaftCommand, Generic[_T]):
        def __init__(self, object_type: type, key: str, index: int, conventions: DocumentConventions):
            if not key:
                raise ValueError("The key must have value")

            super().__init__(CompareExchangeResult[_T])
            self._object_type = object_type
            self._key = key
            self._index = index
            self._conventions = conventions

        def is_read_request(self) -> bool:
            return True

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request(
                "DELETE",
                f"{node.url}/databases/{node.database}/cmpxchg?key={Utils.quote_key(self._key)}&index={self._index}",
            )

        def set_response(self, response: str, from_cache: bool) -> None:
            self.result = CompareExchangeResult.parse_from_string(self._object_type, response, self._conventions)

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class GetCompareExchangeValuesOperation(IOperation[Dict[str, CompareExchangeValue[_T]]], Generic[_T]):
    def __init__(
        self,
        keys: Optional[Collection[str]] = None,
        object_type: Optional[Type[_T]] = None,
        materialize_metadata: Optional[bool] = True,
    ):
        self._materialize_metadata = materialize_metadata
        self._keys = keys
        self._object_type = object_type
        self._start = None
        self._page_size = None
        self._start_with = None

    @classmethod
    def create_for_start_with(
        cls,
        start_with: str,
        start: Optional[int] = None,
        page_size: Optional[int] = None,
        object_type: Optional[Type[_T]] = None,
    ) -> GetCompareExchangeValuesOperation:
        operation = cls(None, object_type, True)
        operation._starts_with = start_with
        operation._start = start
        operation._page_size = page_size
        return operation

    def get_command(self, store: DocumentStore, conventions: DocumentConventions, cache: HttpCache) -> RavenCommand[_T]:
        return self.GetCompareExchangeValuesCommand(self, self._materialize_metadata, conventions)

    class GetCompareExchangeValuesCommand(RavenCommand[Dict[str, CompareExchangeValue[_T]]], Generic[_T]):
        def __init__(
            self,
            operation: GetCompareExchangeValuesOperation[_T],
            materialize_metadata: bool,
            conventions: DocumentConventions,
        ):
            super().__init__(dict)
            self._operation = operation
            self._materialize_metadata = materialize_metadata
            self._conventions = conventions

        def is_read_request(self) -> bool:
            return True

        def create_request(self, node: ServerNode) -> requests.Request:
            path_builder = [node.url, "/databases/", node.database, "/cmpxchg?"]

            if self._operation._keys:
                for key in self._operation._keys:
                    path_builder.append("&key=")
                    path_builder.append(Utils.quote_key(key))
            else:
                if self._operation._start_with:
                    path_builder.append("&startsWith=")
                    path_builder.append(Utils.quote_key(self._operation._start_with))
                if self._operation._start:
                    path_builder.append("&start=")
                    path_builder.append(Utils.quote_key(self._operation._start))
                if self._operation._page_size:
                    path_builder.append("&pageSize=")
                    path_builder.append(Utils.quote_key(self._operation._page_size))

            return requests.Request("GET", "".join(path_builder))

        def set_response(self, response: str, from_cache: bool) -> None:
            self.result = CompareExchangeValueResultParser.get_values(
                self._operation._object_type, response, self._materialize_metadata, self._conventions
            )



from abc import abstractmethod
from typing import Optional


class RavenException(RuntimeError):
    def __init__(self, message: str = None, cause: BaseException = None):
        super(RavenException, self).__init__((message, cause) or message)
        self.reached_leader = None

    @classmethod
    def generic(cls, error: str, json: str):
        return cls(f"{error}. Response: {json}")


class BadResponseException(RavenException):
    def __init__(self, message: str = None, cause: BaseException = None):
        super(BadResponseException, self).__init__(message, cause)


class ConflictException(RavenException):
    @abstractmethod
    def __init__(self, message: str = None, cause: BaseException = None):
        super().__init__(message, cause)


class ConcurrencyException(ConflictException):
    def __init__(self, message):
        super().__init__(message)


class ClientVersionMismatchException(RavenException):
    def __init__(self, message: Optional[str] = None, cause: Optional[Exception] = None):
        super().__init__(message, cause)


class PortInUseException(RavenException):
    pass


class IndexCompactionInProgressException(RavenException):
    pass



from __future__ import annotations

import enum
from typing import Union, List, Dict

from ravendb.primitives import constants


class FacetAggregation(enum.Enum):
    NONE = "None"
    MAX = "Max"
    MIN = "Min"
    AVERAGE = "Average"
    SUM = "Sum"


class FacetOptions:
    @classmethod
    def default_options(cls) -> FacetOptions:
        return cls()

    def __init__(self):
        self.page_size: int = constants.int_max
        self.start: Union[None, int] = None
        self.term_sort_mode: FacetTermSortMode = FacetTermSortMode.VALUE_ASC
        self.include_remaining_terms: bool = False

    def to_json(self) -> dict:
        return {
            "PageSize": self.page_size,
            "Start": self.start or 0,
            "TermSortMode": str(self.term_sort_mode),
            "IncludeRemainingTerms": self.include_remaining_terms,
        }

    @classmethod
    def from_json(cls, json_dict: Dict) -> FacetOptions:
        options = cls()
        options.page_size = json_dict["PageSize"]
        options.start = json_dict["Start"]
        options.term_sort_mode = FacetTermSortMode(json_dict["TermSortMode"])
        options.include_remaining_terms = json_dict["IncludeRemainingTerms"]
        return options


class FacetResult:
    def __init__(self):
        self.name: Union[None, str] = None
        self.values: List[FacetValue] = []
        self.remaining_terms: List[str] = []
        self.remaining_terms_count: Union[None, int] = None
        self.remaining_hits: Union[None, int] = None

    @classmethod
    def from_json(cls, json_dict: dict) -> FacetResult:
        x = cls()
        x.name = json_dict.get("Name", None)
        x.values = [FacetValue.from_json(jv) for jv in json_dict.get("Values", None)]
        x.remaining_terms = json_dict.get("RemainingTerms", None)
        x.remaining_terms_count = json_dict.get("RemainingTermsCount", None)
        x.remaining_hits = json_dict.get("RemainingHits", None)
        return x


class FacetValue:
    def __init__(self, name: str, range_: str, count_: int, sum_: int, max_: int, min_: int, average_: int):
        self.name = name
        self.range_ = range_
        self.count_ = count_
        self.sum_ = sum_
        self.max_ = max_
        self.min_ = min_
        self.average_ = average_

    def __str__(self):
        msg = f"{self.range_} - Count: {self.count_}, "
        if self.sum_ is not None:
            msg += f"Sum: {self.sum_},"
        if self.max_ is not None:
            msg += f"Max: {self.max_},"
        if self.min_ is not None:
            msg += f"Min: {self.min_},"
        if self.sum_ is not None:
            msg += f"Average: {self.average_},"
        if self.name is not None:
            msg += f"Name: {self.name},"

        return msg.removesuffix(";")

    @classmethod
    def from_json(cls, json_dict: dict) -> FacetValue:
        return cls(
            json_dict.get("Name"),
            json_dict.get("Range"),
            json_dict.get("Count"),
            json_dict.get("Sum"),
            json_dict.get("Max"),
            json_dict.get("Min"),
            json_dict.get("Average"),
        )


class FacetTermSortMode(enum.Enum):
    VALUE_ASC = "ValueAsc"
    VALUE_DESC = "ValueDesc"
    COUNT_ASC = "CountAsc"
    COUNT_DESC = "CountDesc"

    def __str__(self):
        return self.value



from typing import Dict, Set, Optional, Callable

from ravendb.primitives import constants
from ravendb.documents.conventions import DocumentConventions
from ravendb.documents.indexes.definitions import (
    FieldStorage,
    FieldIndexing,
    FieldTermVector,
    IndexDefinition,
    IndexSourceType,
    IndexFieldOptions,
    IndexType,
)
from ravendb.documents.indexes.abstract_index_creation_tasks import (
    AbstractIndexCreationTaskBase,
    AbstractIndexDefinitionBuilder,
)
from ravendb.documents.indexes.spatial.configuration import SpatialOptions, SpatialOptionsFactory


class CountersIndexDefinition(IndexDefinition):
    @property
    def source_type(self) -> IndexSourceType:
        return IndexSourceType.COUNTERS


class AbstractGenericCountersIndexCreationTask(AbstractIndexCreationTaskBase[CountersIndexDefinition]):
    def __init__(self):
        super(AbstractGenericCountersIndexCreationTask, self).__init__()
        self._reduce: Optional[str] = None

        self._stores_strings: Dict[str, FieldStorage] = {}
        self._indexes_strings: Dict[str, FieldIndexing] = {}
        self._analyzers_strings: Dict[str, str] = {}
        self._index_suggestions: Set[str] = set()
        self._term_vectors_strings: Dict[str, FieldTermVector] = {}
        self._spatial_options_strings: Dict[str, SpatialOptions] = {}

        self._output_reduce_to_collection: Optional[str] = None
        self._pattern_for_output_reduce_to_collection_references: Optional[str] = None
        self._pattern_references_collection_name: Optional[str] = None

    @property
    def is_map_reduce(self) -> bool:
        return self._reduce is not None

    def _index(self, field: str, indexing: FieldIndexing) -> None:
        self._indexes_strings[field] = indexing

    def _spatial(self, field: str, indexing: Callable[[SpatialOptionsFactory], SpatialOptions]) -> None:
        self._spatial_options_strings[field] = indexing(SpatialOptionsFactory())

    def _store_all_fields(self, storage: FieldStorage) -> None:
        self._stores_strings[constants.Documents.Indexing.Fields.ALL_FIELDS] = storage

    def _store(self, field: str, storage: FieldStorage) -> None:
        self._stores_strings[field] = storage

    def _analyze(self, field: str, analyzer: str) -> None:
        self._analyzers_strings[field] = analyzer

    def _term_vector(self, field: str, term_vector: FieldTermVector) -> None:
        self._term_vectors_strings[field] = term_vector

    def _suggestion(self, field: str) -> None:
        self._index_suggestions.add(field)


class CountersIndexDefinitionBuilder(AbstractIndexDefinitionBuilder[CountersIndexDefinition]):
    def __init__(self, index_name: Optional[str] = None):
        super(CountersIndexDefinitionBuilder, self).__init__(index_name)
        self.map: Optional[str] = None

    def _new_index_definition(self) -> CountersIndexDefinition:
        return CountersIndexDefinition()

    def to_index_definition(
        self, conventions: DocumentConventions, validate_map: bool = True
    ) -> CountersIndexDefinition:
        if self.map is None and validate_map:
            raise RuntimeError(
                f"Map is required to generate an index, you "
                f"cannot create an index without a valid map property (in index {self._index_name})."
            )

        return super(CountersIndexDefinitionBuilder, self).to_index_definition(conventions, validate_map)

    def _to_index_definition(self, index_definition: CountersIndexDefinition, conventions: DocumentConventions) -> None:
        if self.map is None:
            return

        index_definition.maps.add(self.map)


class AbstractCountersIndexCreationTask(AbstractGenericCountersIndexCreationTask):
    def __init__(self):
        super(AbstractCountersIndexCreationTask, self).__init__()
        self.map: Optional[str] = None

    def create_index_definition(self) -> CountersIndexDefinition:
        if self.conventions is None:
            self.conventions = DocumentConventions()

        index_definition_builder = CountersIndexDefinitionBuilder(self.index_name)
        index_definition_builder.indexes_strings = self._indexes_strings
        index_definition_builder.analyzers_strings = self._analyzers_strings
        index_definition_builder.map = self.map
        index_definition_builder.reduce = self._reduce
        index_definition_builder.stores_strings = self._stores_strings
        index_definition_builder.suggestions_options = self._index_suggestions
        index_definition_builder.term_vectors_strings = self._term_vectors_strings
        index_definition_builder.spatial_indexes_strings = self._spatial_options_strings
        index_definition_builder.output_reduce_to_collection = self._output_reduce_to_collection
        index_definition_builder.pattern_for_output_reduce_to_collection_references = (
            self._pattern_for_output_reduce_to_collection_references
        )
        index_definition_builder.pattern_references_collection_name = self._pattern_references_collection_name
        index_definition_builder.additional_sources = self.additional_sources
        index_definition_builder.additional_assemblies = self.additional_assemblies
        index_definition_builder.configuration = self.configuration
        index_definition_builder.lock_mode = self.lock_mode
        index_definition_builder.priority = self.priority
        index_definition_builder.state = self.state
        index_definition_builder.deployment_mode = self.deployment_mode

        return index_definition_builder.to_index_definition(self.conventions)


class AbstractJavaScriptCountersIndexCreationTask(AbstractIndexCreationTaskBase[CountersIndexDefinition]):
    def __init__(self):
        super(AbstractJavaScriptCountersIndexCreationTask, self).__init__()
        self._definition = CountersIndexDefinition()

    @property
    def maps(self) -> Set[str]:
        return self._definition.maps

    @maps.setter
    def maps(self, value: Set[str]):
        self._definition.maps = value

    @property
    def fields(self) -> Dict[str, IndexFieldOptions]:
        return self._definition.fields

    @fields.setter
    def fields(self, value: Dict[str, IndexFieldOptions]):
        self._definition.fields = value

    @property
    def reduce(self) -> str:
        return self._definition.reduce

    @reduce.setter
    def reduce(self, value: Dict[str, IndexFieldOptions]):
        self._definition.reduce = value

    @property
    def is_map_reduce(self) -> bool:
        return self.reduce is not None

    @property
    def output_reduce_to_collection(self) -> str:
        return self._definition.output_reduce_to_collection

    @output_reduce_to_collection.setter
    def output_reduce_to_collection(self, value: Dict[str, IndexFieldOptions]):
        self._definition.output_reduce_to_collection = value

    @property
    def pattern_references_collection_name(self) -> str:
        return self._definition.pattern_references_collection_name

    @pattern_references_collection_name.setter
    def pattern_references_collection_name(self, value: Dict[str, IndexFieldOptions]):
        self._definition.pattern_references_collection_name = value

    @property
    def pattern_for_output_reduce_to_collection_references(self) -> str:
        return self._definition.pattern_for_output_reduce_to_collection_references

    @pattern_for_output_reduce_to_collection_references.setter
    def pattern_for_output_reduce_to_collection_references(self, value: Dict[str, IndexFieldOptions]):
        self._definition.pattern_for_output_reduce_to_collection_references = value

    def create_index_definition(self) -> CountersIndexDefinition:
        self._definition.name = self.index_name
        self._definition.type = IndexType.JAVA_SCRIPT_MAP_REDUCE if self.is_map_reduce else IndexType.JAVA_SCRIPT_MAP
        self._definition.additional_sources = self.additional_sources or {}
        self._definition.additional_assemblies = self.additional_assemblies or set()
        self._definition.configuration = self.configuration
        self._definition.lock_mode = self.lock_mode
        self._definition.priority = self.priority
        self._definition.state = self.state
        self._definition.deployment_mode = self.deployment_mode
        return self._definition



from abc import abstractmethod
from typing import List


class QueryToken:
    RQL_KEYWORDS = {"as", "select", "where", "load", "group", "order", "include"}

    @abstractmethod
    def write_to(self, writer: List[str]) -> None:
        raise NotImplementedError("Error : write_to is not implemented")

    def write_field(self, writer: List[str], field: str):
        key_word = self.is_keyword(field)
        if key_word:
            writer.append("'")
        writer.append(field)
        if key_word:
            writer.append("'")

    @staticmethod
    def is_keyword(field: str) -> bool:
        return field in QueryToken.RQL_KEYWORDS
